#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from fnmatch import fnmatch
import glob
import os
import numpy as np
import simplejson

# internal pyaerocom imports
from pyaerocom._lowlevel_helpers import (check_dirs_exist, dict_to_str)
from pyaerocom import const
from pyaerocom import __version__ as pyaerocom_version
from pyaerocom.mathutils import calc_statistics
from pyaerocom.region import (get_all_default_region_ids, 
                              find_closest_region_coord,
                              Region)
from pyaerocom.io.helpers import save_dict_json

from pyaerocom.web.helpers import (ObsConfigEval, ModelConfigEval, 
                                   update_menu_evaluation_iface, 
                                   make_info_table_evaluation_iface, 
                                   read_json, write_json)
from pyaerocom import ColocationSetup, ColocatedData, Colocator

from pyaerocom.web.obs_config_default import OBS_SOURCES, OBS_DEFAULTS
from pyaerocom.exceptions import DataDimensionError

# ToDo: complete docstring
class AerocomEvaluation(object):
    """Class for creating json files for Aerocom Evaluation interface
    
    High level interface for computation of colocated netcdf files and json
    files for `Aerocom Evaluation interface 
    <https://aerocom-trends.met.no/evaluation/web/>`__. The processing is 
    done *per experiment*. See class attributes for setup options.
    An *experiment* denotes a setup comprising one or more observation 
    networks (specified in :attr:`obs_config`) and the specification of one 
    or more model runs via :attr:`model_config`. These two configuration 
    attributes are dictionaries, where keys correspond to the name of the
    obs / model (as it should appear online) and values specify relevant 
    information for importing the data (see also classes 
    :class:`ModelConfigEval` and :class:`ObsConfigEval`).
    
    In addition to :attr:`model_config, obs_config`, there are more setup 
    settings and options, some of which NEED to be specified (e.g. 
    :attr:`proj_id, exp_id, out_basedir`) and others that may be explicitly 
    specified if desired (e.g. :attr:`harmonise_units, remove_outliers, 
    clear_existing_json`).
    
    The analysis (which can be run via :func:`run_evaluation`) can be summarised
    as follows: for each combination of *variable / obs / model* create one 
    colocated NetCDF file and based on this colocated (single variable) NetCDF 
    file compute all relevant json files for the interface. 
    
    General settings for the colocation (e.g. colocation frequency, start / 
    stop time) can be specified in :attr:`colocation_settings` (for details 
    see :class:`pyaerocom.ColocationSetup`). The colocation routine uses the 
    variables specified in the obs_config entry of the observation network that 
    is supposed to be colocated. If these variables are not provided in a 
    model run or are named differently, then the corresponding model variable 
    that is supposed to be colocated with the observation variable can be 
    specified in the corresponding model entry in :attr:`model_config` via 
    model_use_vars (mapping). Note that this may lead to unit conversion errors 
    if the mapped model variable has a different AeroCom default unit and if 
    outlier ranges are not specified explicitely (see info below).
    
    If :attr:`remove_outliers` is True, then custom outlier ranges for 
    individual variables may be specified in corresponding entry of model in 
    :attr:`model_config`, or, for observation variables
    in :attr:`obs_config`, respectively. 
    NOTE: for each variable that has not specified an outlier range here, the 
    AeroCom default is used, which is specified in pyaerocom data file 
    variables.ini. In the latter case, a unit check is performed in order to 
    make sure the retrieved default outlier ranges are valid. This can cause 
    errors in the colocation if, for instance, a model variable name is used
    that has a different AerocCom default unit than the used observation 
    variable.
    
    Attributes
    ----------
    proj_id : str
        ID of project
    exp_id : str
        ID of experiment
    exp_name : :obj:`str`, optional
        name of experiment 
    clear_existing_json : bool
        Boolean specifying whether existing json files should be deleted for
        var / obs / model combination before rerunning
    out_basedir : str
        basic output directory which defines all further output paths for json
        files 
    out_dirs : dict
        dictionary that specifies the output paths for the different types of
        json files (e.g. map, ts, etc.). Is filled automatically using 
        :attr:`out_basedir, proj_id, exp_id`. Non existing paths are 
        automatically created.
    colocation_settings : dict
        dictionary specifying settings and options for the colocation routine
        (cf. :class:`pyaerocom.ColocationSetup` for available options). 
        Note: the options that are specified in this dictionary are to be 
        understood as global colocation options (for all model / obs 
        combinations defining this experiment). They may be refined or 
        overwritten as required on an individual basis in the definitions for 
        the observations (:attr:`obs_config`) and / or in the model definitions
        (:attr:`model_config`), respectively. The logical order that defines 
        the colocation settings for a certain run (that is, a combination of 
        `var_name, obs_name, model_name`) is:
        is:
            
            1. Import `dict` :attr:`colocation_settings` (global)
            2. Update `dict` with settings from :attr:`obs_config` for \
            `obs_name` (defines `var_name`)
            3. Update `dict` with settings from :attr:`model_config` for \
            `model_name`
     
    add_methods_file : str, optional
        file specifying custom reading methods
    add_methods : dict
        dictionary containing additional reading method
    obs_config : dict
        dictionary containing configuration details for individual observations
        (i.e. instances of :class:`ObsConfigEval` for each observation) used 
        for the analysis.
    obs_ignore : list, optional
        list of observations that are supposed to be ignored in analysis 
        (keys from :attr:`obs_config`)
    model_config : dict
        dictionary containing configuration details for individual models
        (i.e. instances of :class:`ModelConfigEval` for each model) used 
        for the analysis.
    model_ignore : list, optional
        list of models that are supposed to be ignored in analysis 
        (keys from :attr:`model_config`)
    var_mapping : dict
        mapping of variable names for menu in interface
    var_order_menu : list, optional
        order of variables in menu
    """
    OUT_DIR_NAMES = ['map', 'ts', 'scat', 'hm', 'profiles']
    
    
    #: Vertical layer ranges
    VERT_LAYERS = {'0-2km'  :   [0, 2000],
                   '2-5km'  :   [2000, 5000],
                   '5-10km' :   [5000, 10000]}
    
    #: Allowed options for vertical codes
    VERT_CODES = ['Surface', 'Column']
    VERT_CODES.extend(VERT_LAYERS)
    
    #: vertical schemes that may be used for colocation
    VERT_SCHEMES = {'Surface' : 'surface'}
    
    #: Attributes that are ignored when writing setup to json file
    JSON_CFG_IGNORE = ['add_methods', '_log', 'out_dirs']
    
    _OPTS_NAMES_OUTPUT = {
            'clear_existing_json' : 'Delete existing json files before reanalysis',
            'reanalyse_existing'  : 'Reanalyse existing colocated NetCDF files',
            'only_colocation'     : 'Run only colocation (no json files computed)',
            'raise_exceptions'    : 'Raise exceptions if they occur'
    }
                          
    def __init__(self, proj_id=None, exp_id=None, config_dir=None, 
                 try_load_json=True, **settings):
        
        self._log = const.print_log
        
        self.proj_id = proj_id
        
        self.exp_id = exp_id
        
        self.exp_name = None
        
        self.clear_existing_json = True
        
        self.only_colocation = False
        
        #: Base directory for output
        self.out_basedir = None
        
        #: Directory that contains configuration files
        self.config_dir = config_dir
        
        #: Output directories for different types of json files (will be filled
        #: in :func:`init_dirs`)
        self.out_dirs = {}
        
        #: Dictionary specifying default settings for colocation
        self.colocation_settings = ColocationSetup()
        
        self.add_methods_file = None
        self.add_methods = {}
        
        #: Dictionary containing configurations for observations
        self.obs_config = {}
        self.obs_ignore = []
        
        #: Dictionary containing configurations for models
        self.model_config = {}
        self.model_ignore = []
        
        self.var_mapping = {}
        self.var_order_menu = []
        
        self._valid_obs_vars = {}
        if try_load_json and proj_id is not None:
            try:
                self.load_config(self.proj_id, self.exp_id, config_dir)
                const.print_log.warn('Found and imported config file for {} / {}'
                                     .format(self.proj_id, self.exp_id))
            except:
                pass
        self.update(**settings)
        
    
    @property
    def proj_dir(self):
        """Project directory"""
        return os.path.join(self.out_basedir, self.proj_id)
    
    
    @property
    def exp_dir(self):
        """Project directory"""
        return os.path.join(self.proj_dir, self.exp_id)
    
    @property
    def coldata_dir(self):
        """Base directory for colocated data files"""
        return self.colocation_settings['basedir_coldata']
    
    @property
    def regions_file(self):
        """json file containing region specifications"""
        return os.path.join(self.out_basedir, 'regions.json')
    
    @property
    def menu_file(self):
        """json file containing region specifications"""
        return os.path.join(self.proj_dir, 'menu.json')
    
    @property
    def all_model_names(self):
        """List of all model names"""
        return list(self.model_config)
    
    @property
    def all_obs_names(self):
        """List of all obs names"""
        return list(self.obs_config)
    
    @property
    def all_map_files(self):
        """List of all existing map files"""
        if not os.path.exists(self.out_dirs['map']):
            raise FileNotFoundError('No data available for this experiment')
        return os.listdir(self.out_dirs['map'])
    
    def _update_custom_read_methods(self):
        for mcfg in self.model_config.values():
            if not 'model_read_aux' in mcfg:
                continue
            maux = mcfg['model_read_aux']
            if maux is None:
                continue
            elif not isinstance(maux, dict):
                raise ValueError('Require dict, got {}'.format(maux))
            for varcfg in maux.values():
                err_msg_base = ('Invalid definition of model_read_aux')
                if not isinstance(varcfg, dict):
                    raise ValueError('{}: value needs to be dictionary'
                                     .format(err_msg_base))
                if not all([x in varcfg for x in ['vars_required', 'fun']]):
                    raise ValueError('{}: require specification of keys '
                                     'vars_required and fun'
                                     .format(err_msg_base))
                if not isinstance(varcfg['fun'], str):
                    raise ValueError('Names of custom methods need to be strings')
                    
                name = varcfg['fun']
                fun = self.get_custom_read_method_model(name)
                if not name in self.add_methods:
                    self.add_methods[name] = fun
                
                    
    def get_custom_read_method_model(self, method_name):
        """Get custom read method for computation of model variables during read
        
        Parameters
        ----------
        method_name : str
            name of method
            
        Returns
        -------
        callable
            corresponding python method
        
        Raises
        ------
        ValueError
            if no method with the input name can be accessed
        """
        if method_name in self.add_methods:
            fun = self.add_methods[method_name]
        else:
            import sys, importlib
            fp = self.add_methods_file
            
            if fp is None or not os.path.exists(fp):
                raise ValueError('Failed to access custom read method {}'
                                 .format(method_name))
            try:
                moddir = os.path.dirname(fp)
                if not moddir in sys.path:
                    sys.path.append(moddir)
                modname = os.path.basename(fp).split('.')[0]
                if '.' in modname:
                    raise NameError('Invalid name for module: {} (file name must '
                                    'not contain .)'.format(fp))
                mod = importlib.import_module(modname)
            except Exception as e:
                raise ImportError('Failed to import module containing '
                                  'additional custom model read methods '
                                  '.Error: {}'.format(repr(e)))
            if not method_name in mod.FUNS:
                raise ValueError('File {} does not contain custom read '
                                 'method: {}'.format(fp, method_name))
            fun = mod.FUNS[method_name]
        #fun = self.add_methods[name]
        if not callable(fun):
            raise TypeError('{} ({}) is not a callable object'.format(fun, 
                            method_name))
        return fun
        
    def update(self, **settings):
        """Update current setup"""
        for k, v in settings.items():
            self[k] = v
        self.check_config()
        self.init_dirs()
        self._update_custom_read_methods()
        
    def _set_obsconfig(self, val):
        cfg = {}
        for k, v in val.items():
            cfg[k] = ObsConfigEval(**v)
        
        self.obs_config = cfg
        
    def _set_modelconfig(self, val):
        cfg = {}
        for k, v in val.items():
            cfg[k] = ModelConfigEval(**v)
        self.model_config = cfg
        self._update_custom_read_methods()
    
            
    def __setitem__(self, key, val):
        if key in self.colocation_settings:
            self.colocation_settings[key] = val
        elif key == 'obs_config':
            self._set_obsconfig(val)
        elif key == 'model_config':
            self._set_modelconfig(val)
        elif key == 'colocation_settings':
            self.colocation_settings.update(**val)
        elif isinstance(key, str) and isinstance(val, dict):
            if 'obs_id' in val:
                if key in self.obs_config:
                    self._log.warn('Obs config for key {}  already exists and '
                                   'will be overwritten {}'.format(key))
                self.obs_config[key] = ObsConfigEval(**val)
            elif 'model_id' in val:
                if key in self.model_config:
                    self._log.warn('Model config for key {}  already exists and '
                                   'will be overwritten {}'.format(key))
                self.model_config[key] = ModelConfigEval(**val)
            else:
                self.__dict__[key] = val
        elif key in self.__dict__:
            self.__dict__[key] = val
        else:
            raise KeyError('Invalid input key {}. Cannot assign {}'
                           .format(key, val))
                
    def __getitem__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        elif key in self.colocation_settings:
            return self.colocation_settings[key]
    
    def init_dirs(self, out_basedir=None):
        """Check and create directories"""
        if out_basedir is not None:
            self.out_basedir = out_basedir
        if self.out_basedir is None:
            self.out_basedir = const.OUTPUTDIR
        check_dirs_exist(self.out_basedir, self.proj_dir, self.exp_dir)
        outdirs = {}
        for dname in self.OUT_DIR_NAMES:
            outdirs[dname] = os.path.join(self.exp_dir, dname)
        check_dirs_exist(**outdirs)
        self.out_dirs = outdirs
    
    
    
    def check_config(self):
        if not isinstance(self.proj_id, str):
            raise ValueError('proj_id must be string, got {}'.format(self.proj_id))
        
        if not isinstance(self.exp_id, str):
            raise ValueError('exp_id must be string, got {}'.format(self.exp_id))
            
        if not isinstance(self.exp_name, str):
            const.print_log.warning('exp_name must be string, got {}. Using '
                                    'exp_id {} for experiment name'
                                    .format(self.exp_name, self.exp_id))
            self.exp_name = self.exp_id
        
        for k, v in self.model_config.items():
            if '_' in k or ':' in k:
                raise NameError('Model config name must not contain _ (underscore) or colon.')
            elif len(k) > 20:
                print('Long model ID: {}. Consider renaming'.format(k))
            elif len(k) > 25:
                raise ValueError('Too long model ID: {} (max 20 chars)'.format(k))
            elif not 'model_id' in v:
                raise KeyError('Model configuration for {} does not contain '
                               'model_id'.format(k))
        for k, v in self.obs_config.items():
            if '_' in k or ':' in k:
                raise NameError('Obs config name must not contain _ (underscore) or colon')
            elif len(k) > 15:
                print('Long obs ID: {}. Consider renaming'.format(k))
            elif len(k) > 20:
                raise ValueError('Too long obs ID: {} (max 20 chars)'.format(k))
            elif not 'obs_id' in v:
                raise KeyError('Obs configuration for {} does not contain '
                               'obs_id'.format(k))
                
    def get_model_id(self, model_name):
        """Get AeroCom ID for model name
        """
        for name, info in self.model_config.items():
            if name == model_name:
                return info['model_id']
        raise KeyError('Cannot find setup for ID {}'.format(model_name))
    
    def get_obs_id(self, obs_name):
        """Get AeroCom ID for obs name
        """
        for name, info in self.obs_config.items():
            if name == obs_name:
                return info['obs_id']
        raise KeyError('Cannot find setup for ID {}'.format(obs_name))
    
    def find_obs_name(self, obs_id, obs_var):
        """Find web menu name of obs dataset based on obs_id and variable
        """
        matches = []
        for obs_name, info in self.obs_config.items():
            if info['obs_id'] == obs_id and obs_var in info['obs_vars']:
                matches.append(obs_name)
        if len(matches) == 1:
            return matches[0]
        raise ValueError('Could not identify unique obs name')
    
    def find_model_name(self, model_id):
        """Find web menu name of model dataset based on model_id
        """
        matches = []
        for model_name, info in self.model_config.items():
            if info['model_id'] == model_id:
                matches.append(model_name)
        if len(matches) == 1:
            return matches[0]
        raise ValueError('Could not identify unique model name')
            
    def make_regions_json(self):
        """Creates file regions.ini for web interface"""
        regs = {}
        for regname in get_all_default_region_ids():
            reg = Region(regname)
            regs[regname] = r = {}
            latr = reg.lat_range
            r['minLat'] = latr[0]
            r['maxLat'] = latr[1]
            lonr = reg.lon_range
            r['minLon'] = lonr[0]
            r['maxLon'] = lonr[1]
        save_dict_json(regs, self.regions_file)
        return regs
    
    def compute_json_files_from_colocateddata(self, coldata, obs_name=None, 
                                              model_name=None):
        """Creates all json files for one ColocatedData object"""
        if not isinstance(coldata, ColocatedData):
            raise ValueError('Need ColocatedData object, got {}'
                             .format(type(coldata)))
        stats_dummy = {}
        
        for k in calc_statistics([1], [1]):
            stats_dummy[k] = np.nan
        
        stacked = False
        if 'altitude' in coldata.data.dims:
            raise NotImplementedError('Cannot yet handle profile data')
        if not 'station_name' in coldata.data.coords:
            if not coldata.data.ndim == 4:
                raise DataDimensionError('Invalid number of dimensions. '
                                         'Need 4, got: {}'
                                         .format(coldata.data.dims))
            elif not 'latitude' in coldata.data.dims and 'longitude' in coldata.data.dims:
                raise DataDimensionError('Need latitude and longitude '
                                         'dimension. Got {}'
                                         .format(coldata.data.dims))
            coldata.data = coldata.data.stack(station_name=('latitude', 
                                                            'longitude'))
            stacked = True
            
        ts_types_order = const.GRID_IO.TS_TYPES
        to_ts_types = ['daily', 'monthly', 'yearly']
        
        data_arrs = dict.fromkeys(to_ts_types)
        jsdate = dict.fromkeys(to_ts_types)
        
        ts_type = coldata.meta['ts_type']
        for freq in to_ts_types:
            if ts_types_order.index(freq) < ts_types_order.index(ts_type):
                data_arrs[freq] = None
            elif ts_types_order.index(freq) == ts_types_order.index(ts_type):
                data_arrs[freq] = coldata.data
                
                js = (coldata.data.time.values.astype('datetime64[s]') - 
                      np.datetime64('1970', '[s]')).astype(int) * 1000
                jsdate[freq] = js.tolist()
                
            else:
                colstp = self.colocation_settings
                _a = coldata.resample_time(to_ts_type=freq,
                                     apply_constraints=colstp.apply_time_resampling_constraints, 
                                     min_num_obs=colstp.min_num_obs,
                                     colocate_time=colstp.colocate_time,
                                     inplace=False).data
                data_arrs[freq] = _a #= resample_time_dataarray(arr, freq=freq)
                js = (_a.time.values.astype('datetime64[s]') - 
                      np.datetime64('1970', '[s]')).astype(int) * 1000
                jsdate[freq] = js.tolist()      
        
        #print(jsdate)
    
        obs_id = coldata.meta['data_source'][0]
        model_id = coldata.meta['data_source'][1]
        
        obs_var = coldata.meta['var_name'][0]
        model_var = coldata.meta['var_name'][1]
        
        if obs_name is None:
            obs_name = self.find_obs_name(obs_id, obs_var)
            
        if model_name is None:
            model_name = self.find_model_name(model_id)
    
        ts_objs = []
        
        map_data = []
        scat_data = {}
        hm_data = {}
        
        
        #ts_type = coldata.meta['ts_type']
        vert_code = self.get_vert_code(obs_name, obs_var)
        
        # data used for heatmap display in interface
        if stacked:    
            hmd = ColocatedData(data_arrs['monthly'].unstack('station_name'))
        else:
            hmd = ColocatedData(data_arrs['monthly'])
# =============================================================================
#         if ts_type == 'monthly':
#             hmd = coldata
#         else:
#             cs = self.colocation_settings
#             ac = cs.apply_time_resampling_constraints
#             if ac is None:
#                 ac = const.OBS_APPLY_TIME_RESAMPLE_CONSTRAINTS
#                 
#             if ac:
#                 mo = cs.min_num_obs
#                 if cs.min_num_obs is None:
#                     mo = const.OBS_MIN_NUM_RESAMPLE
#                 
#             hmd  = coldata.resample_time(to_ts_type='monthly',
#                                    apply_constraints=ac,
#                                    min_num_obs=mo,
#                                    colocate_time=cs['colocate_time'], 
#                                    inplace=False)
#                                    
# =============================================================================
        for reg in get_all_default_region_ids():
            filtered = hmd.apply_latlon_filter(region_id=reg)
            stats = filtered.calc_statistics()
            for k, v in stats.items():
                stats[k] = np.float64(v)
            
            hm_data[reg] = stats
        
        self._write_heatmap_json(hm_data, obs_name, obs_var, vert_code, 
                                 model_name, model_var)
        
        if vert_code == 'ModelLevel':
            raise NotImplementedError('Coming soon...')
        const.print_log.info('Computing json files for {} vs. {}'
                             .format(model_name, obs_name))
        for i, stat_name in enumerate(coldata.data.station_name.values):
            _disp = ('{} - {} ({}) vs. {} ({})'
                     .format(stat_name, model_name, 
                             coldata.meta['var_name'][1],
                             obs_name, coldata.meta['var_name'][0]))
            has_data = False
            ts_data = {}
            ts_data['station_name'] = stat_name
            ts_data['pyaerocom_version'] = pyaerocom_version
            ts_data['obs_name'] = obs_name
            ts_data['model_name'] = model_name
            ts_data['obs_var'] = coldata.meta['var_name'][0]
            ts_data['obs_unit'] = coldata.meta['var_units'][0]
            ts_data['vert_code'] = vert_code
            ts_data['obs_freq_src'] = coldata.meta['ts_type_src'][0]
            ts_data['obs_revision'] = coldata.meta['revision_ref']
            
            ts_data['mod_var'] = coldata.meta['var_name'][1]
            ts_data['mod_unit'] = coldata.meta['var_units'][1]
            ts_data['mod_freq_src'] = coldata.meta['ts_type_src'][1]
            
            stat_lat = np.float64(coldata.data.latitude[i])
            stat_lon = np.float64(coldata.data.longitude[i])
            if 'altitude' in coldata.data.coords:
                stat_alt = np.float64(coldata.data.altitude[i])
            else:
                stat_alt = np.nan
            region = find_closest_region_coord(stat_lat, stat_lon)
            
            # station information for map view
            map_stat = {'site'      : stat_name, 
                        'lat'       : stat_lat, 
                        'lon'       : stat_lon,
                        'alt'       : stat_alt,
                        'region'    : region}
            
            for tres, arr in data_arrs.items():
                map_stat['{}_statistics'.format(tres)] = {}
                if arr is None:
                    ts_data['{}_date'.format(tres)] = []
                    ts_data['{}_obs'.format(tres)] = []
                    ts_data['{}_mod'.format(tres)] = []
                    map_stat['{}_statistics'.format(tres)].update(stats_dummy)
                    continue
        
                obs_vals = arr.sel(data_source=obs_id, 
                                   station_name=stat_name).values

                if all(np.isnan(obs_vals)):
                    _disp += ': No obs data'
                    ts_data['{}_date'.format(tres)] = []
                    ts_data['{}_obs'.format(tres)] = []
                    ts_data['{}_mod'.format(tres)] = []
                    map_stat['{}_statistics'.format(tres)].update(stats_dummy)
                    continue
                has_data = True
                mod_vals = arr.sel(data_source=model_id, 
                                   station_name=stat_name).values
                
                if not len(jsdate[tres]) == len(obs_vals):
                    raise Exception('Please debug...')
                
                ts_data['{}_date'.format(tres)] = jsdate[tres]
                ts_data['{}_obs'.format(tres)] = obs_vals.tolist()
                ts_data['{}_mod'.format(tres)] = mod_vals.tolist()
                
                station_statistics = calc_statistics(mod_vals, obs_vals)
                for k, v in station_statistics.items():
                    station_statistics[k] = np.float64(v)
                map_stat['{}_statistics'.format(tres)] = station_statistics
            
            if has_data:
                _disp += ': OK'
                ts_objs.append(ts_data)
                map_data.append(map_stat)
                scat_data[str(stat_name)] = sc = {}
                sc['obs'] = ts_data['monthly_obs']
                sc['mod'] = ts_data['monthly_mod']
                sc['region'] = region
            const.print_log.info(_disp)
        dirs = self.out_dirs
    
        map_name = self.get_json_mapname(obs_name, obs_var, model_name, 
                                         model_var, vert_code)
        
        outfile_map =  os.path.join(dirs['map'], map_name)
        with open(outfile_map, 'w') as f:
            simplejson.dump(map_data, f, ignore_nan=True)
        
        outfile_scat =  os.path.join(dirs['scat'], map_name)
        with open(outfile_scat, 'w') as f:
            simplejson.dump(scat_data, f, ignore_nan=True)
            
        for ts_data in ts_objs:
            #writes json file
            self._write_stationdata_json(ts_data)
         
    def get_vert_code(self, obs_name, obs_var):
        """Get vertical code name for obs / var combination"""
        info =  self.obs_config[obs_name]['obs_vert_type']
        if isinstance(info, str):
            return info
        return info[obs_var]
    
    @property
    def _heatmap_file(self):
        return os.path.join(self.out_dirs['hm'], 'glob_stats.json')
    
    def update_heatmap_json(self):
        fp = self._heatmap_file
        if not os.path.exists(fp):
            raise FileNotFoundError(fp)
            
        with open(self.menu_file, 'r') as f:
            menu = simplejson.load(f)
        with open(fp, 'r') as f:
            data = simplejson.load(f)
        if not self.exp_id in menu:
            raise ValueError('No entry found in menu.json for experiment {}'
                             .format(self.exp_id))
        
        menu = menu[self.exp_id]
        hm = {}
        for var, info in menu.items():
            obs_dict = info['obs']
            if not var in hm:
                hm[var] = {}
            for obs, vdict in obs_dict.items():
                if not obs in hm[var]:
                    hm[var][obs] = {}
                for vc, mdict in vdict.items():
                    if not vc in hm[var][obs]:
                        hm[var][obs][vc] = {}
                    for mod, minfo in mdict.items():
                        if not mod in hm[var][obs][vc]:
                            hm[var][obs][vc][mod] = {}
                        modvar = minfo['var']
                        if not modvar in hm[var][obs][vc][mod]:
                            hm[var][obs][vc][mod][modvar] = {}
                        hm_data = data[var][obs][vc][mod][modvar]
                        hm[var][obs][vc][mod][modvar] = hm_data
                            
                            
# =============================================================================
#         new = sort_dict_by_name(data, pref_list=self.var_order_menu)
#         new_sorted = {}
#         for var, obs_dict in new.items():
#             sorted_obs = sort_dict_by_name(obs_dict)
#             new_sorted[var] = sorted_obs
#             for obs_name, vert_codes in sorted_obs.items():
#                 vert_codes_sorted = sort_dict_by_name(vert_codes)
#                 new_sorted[var][obs_name] = vert_codes_sorted
#                 for vert_code, models in vert_codes_sorted.items():
#                     models_sorted = sort_dict_by_name(models)
#                     new_sorted[var][obs_name][vert_code] = models_sorted
# =============================================================================
        with open(fp, 'w') as f:
            simplejson.dump(hm, f, ignore_nan=True)
        
    def _write_heatmap_json(self, result, obs_name, obs_var, vert_code, 
                            model_name, model_var):
        fp = self._heatmap_file
        if os.path.exists(fp):
            try:
                with open(fp, 'r') as f:
                    current = simplejson.load(f)
            except Exception as e:
                raise Exception('Fatal: could not open existing json file: {}. '
                                'Reason: {}'.format(fp, repr(e)))
        else:
            current = {}
        if not obs_var in current:
            current[obs_var] = {}
        ov = current[obs_var]
        if not obs_name in ov:
            ov[obs_name] = {}
        on = ov[obs_name]
        if not vert_code in on:
            on[vert_code] = {}
        ovc = on[vert_code]
        if not model_name in ovc:
            ovc[model_name] = {}
        mn = ovc[model_name]
        if model_var in mn:
            const.print_log.info('Overwriting existing heatmap statistics for '
                                 'model {}/{} ({}, {}, {}) in glob_stats.json'
                                 .format(model_name, model_var, obs_var, obs_name, 
                                         vert_code))
        mn[model_var] = result
        with open(fp, 'w') as f:
            simplejson.dump(current, f, ignore_nan=True)
        
    def _write_stationdata_json(self, ts_data):
        
        filename = self.get_stationfile_name(ts_data['station_name'], 
                                             ts_data['obs_name'],
                                             ts_data['obs_var'],
                                             ts_data['vert_code'])
    
        fp = os.path.join(self.out_dirs['ts'], filename)
        if os.path.exists(fp):
            try:
                with open(fp, 'r') as f:
                    current = simplejson.load(f)
            except Exception as e:
                raise Exception('Fatal: could not open existing json file: {}. '
                                'Reason: {}'.format(fp, repr(e)))
        else:
            current = {}
        current[ts_data['model_name']] = ts_data
        with open(fp, 'w') as f:
            simplejson.dump(current, f, ignore_nan=True)
            
    def get_stationfile_name(self, station_name, obs_name, obs_var, vert_code):
        """Get name of station timeseries file"""
        return ('{}_OBS-{}:{}_{}.json'
                .format(station_name, obs_name, obs_var, vert_code))
    
    
    def get_json_mapname(self, obs_name, obs_var, model_name, model_var, 
                           vert_code):
        """Get name base name of json file""" 
        return ('OBS-{}:{}_{}_MOD-{}:{}.json'
                .format(obs_name, obs_var, vert_code, model_name, model_var))
    
    def find_coldata_files(self, model_name, obs_name, var_name=None):
        """Find colocated data files for a certain model/obs/var combination
        
        Parameters
        ----------
        model_name : str
            name of model
        obs_name : str
            name of observation network
        var_name : str, optional
            name of variable.
            
        Returns
        -------
        list
            list of file paths of ColocatedData files that match input specs
        """
        
        files = []
        model_id = self.get_model_id(model_name)
        coldata_dir = os.path.join(self.coldata_dir, model_id)
        if os.path.exists(coldata_dir):
            for fname in os.listdir(coldata_dir):
                try:
                    m = ColocatedData.get_meta_from_filename(fname)
                    match = (m['data_source'][0] == obs_name and
                             m['data_source'][1] == model_name)
                    if match:
                        files.append(os.path.join(coldata_dir, fname))
                except:
                    const.print_log.warning('Invalid file {} in coldata dir'
                                            .format(fname))

        if len(files) == 0:
            msg = ('Could not find any colocated data files for model {}, '
                   'obs {}'
                   .format(model_name, obs_name))
            if self.colocation_settings['raise_exceptions']:
                raise IOError(msg)
            else:
                self._log.warning(msg)
        return files
    
    def make_json_files(self, model_name, obs_name, var_name=None,
                        colocator=None):
        """Convert colocated data file(s) in model data directory into json
        
        Parameters
        ----------
        model_name : str
            name of model run
        obs_name : str
            name of observation network
        var_name : str, optional
            name of variable supposed to be analysed. If None, then all
            variables available for observation network are used (defined in 
            :attr:`obs_config` for each entry). Defaults to None.
        colocator : Colocator, optional
            instance of colocator class containing information about which
            files to process (e.g. created in :func:`run_evaluation`). If None,
            than all colocated data files are processed that are located in the
            corresponding colocation data directory and that match the input 
            specs.
        
        Returns
        -------
        list
            list of colocated data files that were converted
        """
        converted = []
        
        files = self.find_coldata_files(model_name, obs_name, var_name)
        for file in files:
            fname = os.path.basename(file)
            if not fname in colocator.file_status:
                const.print_log.info('Skipping computation of json files from '
                                     'colocated data file {}. This file is not '
                                     'part of this experiment (obs config)'
                                     .format(fname, self.obs_config[obs_name]))
                continue
            if colocator.file_status[fname] == 'skipped':
                const.print_log.info('Recomputing json files for existing '
                                     'colocated data file')
            elif not colocator.file_status[fname] == 'saved':
                const.print_log.info('Skipping computation of json files from '
                                     'colocated data file {}. Colocator object '
                                     'has marked this file as {} (need either '
                                     'status skipped or saved)'
                                     .format(fname, colocator.file_status[fname]))
                continue
            const.print_log.info('Processing file {}'.format(file))
            d = ColocatedData(file)
            self.compute_json_files_from_colocateddata(d, obs_name, model_name)
            converted.append(file)
        return converted
    
    def delete_all_colocateddata_files(self, model_name, obs_name):
        #model_id = self.get_model_id(model_name)
        #obs_id = self.get_obs_id(obs_name)
        obs_vars = self.obs_config[obs_name]['obs_vars']
    
        for obs_var in obs_vars:
            files = glob.glob('{}/{}/{}*REF-{}*.nc'
                              .format(self.coldata_dir, model_name, obs_var, 
                                      obs_name))
            for file in files:
                const.print_log.info('DELETING FILE: {}'.format(file))
                os.remove(file)
            
    def run_colocation(self, model_name, obs_name, var_name=None):
        """Run colocation for model / obs combination
        
        Parameters
        ----------
        model_name : str or list, optional
            Name or pattern specifying model that is supposed to be analysed.
            Can also be a list of names or patterns to specify multiple models.
            If None (default), then all models are run that are part of this
            experiment.
        obs_name : :obj:`str`, or :obj:`list`, optional
            Like :attr:`model_name`, but for specification(s) of observations
            that are supposed to be used. If None (default) all observations 
            are used.
        var_name : str, optional
            name of variable supposed to be analysed. If None, then all
            variables available for observation network are used (defined in 
            :attr:`obs_config` for each entry). Defaults to None.
            
        Returns
        -------
        Colocator
            instance of colocation class
        """
        if self.colocation_settings['reanalyse_existing']:
            self.delete_all_colocateddata_files(model_name, obs_name)
            
        col = Colocator()
        col.update(**self.colocation_settings)
        
        if not model_name in self.model_config:
            raise KeyError('No such model name in configuration: {}. Available '
                           'names: {}'.format(model_name, 
                                              self.all_model_names))
        elif not obs_name in self.obs_config:
            raise KeyError('No such obs name in configuration: {}. Available '
                           'names: {}'.format(obs_name, self.all_obs_names))
            
        obs_cfg = self.obs_config[obs_name]
        if obs_cfg['obs_vert_type'] in self.VERT_SCHEMES and not 'vert_scheme' in obs_cfg:
            obs_cfg['vert_scheme'] = self.VERT_SCHEMES[obs_cfg['obs_vert_type']]
        col.update(**obs_cfg)
        col.update(**self.get_model_config(model_name))
        
        const.print_log.info('Running colocation of {} against {}'
                             .format(model_name, obs_name))
        # for specifying the model and obs names in the colocated data file
        col.model_name = model_name
        col.obs_name = obs_name
        
        # run colocation
        col.run(var_name)
        
        return col
    
    def get_model_config(self, model_name):
        """Get model configuration
        
        Since the configuration files for experiments are in json format, they
        do not allow the storage of executable custom methods for model data
        reading. Instead, these can be specified in a python module that may
        be specified via :attr:`add_methods_file` and that contains a 
        dictionary `FUNS` that maps the method names with the callable methods.
        
        As a result, this means that, by default, custom read methods for 
        individual models in :attr:`model_config` do not contain the 
        callable methods but only the names. This method will take care of 
        handling this and will return a dictionary where potential custom 
        method strings have been converted to the corresponding callable 
        methods.
        
        Parameters
        ----------
        model_name : str
            name of model run
            
        Returns 
        -------
        dict
            Dictionary that specifies the model setup ready for the analysis
        """
        mcfg = self.model_config[model_name]
        outcfg = {}
        if not 'model_id' in mcfg:
            raise ValueError('Model configuration for {} is missing '
                             'specification of model_id '.format(model_name))
        for key, val in mcfg.items():
            
            if key != 'model_read_aux':
                outcfg[key] = val
            else:
                outcfg[key] = d = {}
                for var, rcfg in val.items():
                    d[var] = {}
                    d[var]['vars_required'] = rcfg['vars_required']
                    fun_str = rcfg['fun']
                    if not isinstance(fun_str, str):
                        raise Exception('Unexpected error. Custom method defs. '
                                        'need to be strings, got {}'.format(fun_str))
                    d[var]['fun'] = self.get_custom_read_method_model(fun_str)
        return outcfg
            
    def find_model_matches(self, name_or_pattern):
        """Find model names that match input search pattern(s)
        
        Parameters
        ----------
        name_or_pattern : :obj:`str`, or :obj:`list`
            Name or pattern specifying model search string. Can also be a list 
            of names or patterns to search for multiple models.
            
        Returns
        -------
        list
            list of model names (i.e. keys of :attr:`model_config`) that match 
            the input search string(s) or pattern(s)
            
        Raises
        ------
        KeyError
            if no matches can be found
        """
        
        
        if isinstance(name_or_pattern, str):
            name_or_pattern = [name_or_pattern]
        from fnmatch import fnmatch
        matches = []
        for search_pattern in name_or_pattern:
            for mname in self.model_config:
                if fnmatch(mname, search_pattern) and not mname in matches:
                    matches.append(mname)
        if len(matches) == 0:
            raise KeyError('No models could be found that match input {}'
                           .format(name_or_pattern))
        return matches
    
    def find_obs_matches(self, name_or_pattern):
        """Find model names that match input search pattern(s)
        
        Parameters
        ----------
        name_or_pattern : :obj:`str`, or :obj:`list`
            Name or pattern specifying obs search string. Can also be a list 
            of names or patterns to search for multiple obs networks.
            
        Returns
        -------
        list
            list of model names (i.e. keys of :attr:`obs_config`) that match 
            the input search string(s) or pattern(s)
            
        Raises
        ------
        KeyError
            if no matches can be found
        """
        
        
        if isinstance(name_or_pattern, str):
            name_or_pattern = [name_or_pattern]
        matches = []
        for search_pattern in name_or_pattern:
            for mname in self.obs_config:
                if fnmatch(mname, search_pattern) and not mname in matches:
                    matches.append(mname)
        if len(matches) == 0:
            raise KeyError('No observations could be found that match input {}'
                           .format(name_or_pattern))
        return matches
    
    def run_evaluation(self, model_name=None, obs_name=None, var_name=None, 
                       update_interface=True, 
                       reanalyse_existing=None, raise_exceptions=None, 
                       clear_existing_json=None, only_colocation=None):
        """Create colocated data and json files for model / obs combination
        
        Parameters
        ----------
        model_name : str or list, optional
            Name or pattern specifying model that is supposed to be analysed.
            Can also be a list of names or patterns to specify multiple models.
            If None (default), then all models are run that are part of this
            experiment.
        obs_name : :obj:`str`, or :obj:`list`, optional
            Like :attr:`model_name`, but for specification(s) of observations
            that are supposed to be used. If None (default) all observations 
            are used.
        var_name : str, optional
            name of variable supposed to be analysed. If None, then all
            variables available for observation network are used (defined in 
            :attr:`obs_config` for each entry). Defaults to None.
        update_interface : bool
            if true, relevant json files that determine what is displayed 
            online are updated after the run, including the the menu.json file 
            and also, the model info table (minfo.json) file is created and
            saved in :attr:`exp_dir`.
        reanalyse_existing : Bool, optional
            if True, existing colocated data files are ignored. If None, the
            class default will be used (defined in config file)
        raise_exceptions : Bool, optional
            if True, exceptions during colocation will be raised if they occur 
            (for debugging). If None, the class default will be used (defined 
            in config file)
        clear_existing_json : Bool, optional
            if True, existing json files for model / obs combination will be 
            deleted before rerun. If None, the
            class default will be used (defined in config file)
        only_colocation : Bool, optional
            if True, only colocation will be performed and no json files will
            be created. If None, the class default will be used (defined in 
            config file)
            
        Returns
        -------
        list
            list containing all colocated data objects that have been converted
            to json files.
        """
        res = None
        if reanalyse_existing is not None:
            self.colocation_settings['reanalyse_existing'] = reanalyse_existing
        if raise_exceptions is not None:
            self.colocation_settings['raise_exceptions'] = raise_exceptions
        if clear_existing_json is not None:
            self.clear_existing_json = clear_existing_json
        if only_colocation is not None:
            self.only_colocation = only_colocation
        
        if self.clear_existing_json:
            self.clean_json_files()
            
        if model_name is None:
            model_list = list(self.model_config)
        else:
            model_list = self.find_model_matches(model_name)
        
        if obs_name is None:
            obs_list = list(self.obs_config)
        else:
            obs_list = self.find_obs_matches(obs_name)
        self._log.info(self.info_string_evalrun(obs_list, model_list))
        
        self._update_custom_read_methods()
        
        for obs_name in obs_list:
            if obs_name in self.obs_ignore:
                self._log.info('Skipping observation {}'.format(obs_name))
                continue
            for model_name in model_list:
                if model_name in self.model_ignore:
                    self._log.info('Skipping model {}'.format(model_name))
                    continue
            
                col = self.run_colocation(model_name, obs_name, var_name)
                if only_colocation:
                    self._log.info('Skipping computation of json files for {}'
                                   '/{}'.format(obs_name, model_name))
                    continue
                res = self.make_json_files(model_name, obs_name, var_name,
                                           colocator=col)
        
        if update_interface:
            #self.clean_json_files()
            self.update_interface()

        return res            
        
    def info_string_evalrun(self, obs_list, model_list):
        """Short information string that summarises settings for evaluation run
        
        Parameters
        ----------
        obs_list
            list of observation names supposed to be processed
        model_list : list
            list of model names supposed to be processed
            
        Returns
        -------
        str
            info string
        """
        s = ('\nRunning analysis:\n'
             'Obs. names: {}\n'
             'Model names: {}\n'
             'Remove outliers: {}\n'
             'Harmonise units: {}'
             .format(obs_list, model_list, self['remove_outliers'], 
                     self['harmonise_units']))
        for k, i in self._OPTS_NAMES_OUTPUT.items():
            s += '\n{}: {}'.format(i, self[k]) 
        s += '\n'
        return s
    
    def check_read_model(self, model_name, var_name, **kwargs):
        if not model_name in self.model_config:
            raise ValueError('No such model available {}'.format(model_name))
        mcfg = self.get_model_config(model_name)
        
        from pyaerocom.io import ReadGridded
        r = ReadGridded(mcfg['model_id'])
        try:
            var_name = mcfg['model_use_vars'][var_name]
        except:
            pass
        
        if 'model_read_aux' in mcfg and var_name in mcfg['model_read_aux']:
            aa = mcfg['model_read_aux'][var_name]
            
        
            r.add_aux_compute(var_name, vars_required=aa['vars_required'], 
                              fun=aa['fun'])
        return r.read_var(var_name, **kwargs)
        
    def _info_from_map_file(self, filename):
        f = filename
        obs_info = f.split('OBS-')[1].split('_MOD')[0].split('_')
        obs_name, obs_var = obs_info[0].split(':')
        vert_code = obs_info[1]
        mod_name, mod_var =  f.split('MOD-')[1].split('.json')[0].split(':')
        return (obs_name, obs_var, vert_code, mod_name, mod_var)
    
    def get_web_overview_table(self):
        """Computes overview table based on existing map files"""
        tab = []
        from pandas import DataFrame
        for f in self.all_map_files:
            if f.endswith('.json') and f.startswith('OBS'):
                (obs_name, obs_var, 
                 vert_code, 
                 mod_name, mod_var) = self._info_from_map_file(f)
                if mod_name in self.model_ignore:
                    continue
                elif not mod_name in self.model_config:
                    const.print_log.warning('Found outdated json map file: {}'
                                            'Will be ignored'.format(f))
                    continue
                elif not obs_name in self.obs_config:
                    const.print_log.warning('Found outdated json map file: {}'
                                            'Will be ignored'.format(f))
                    continue
                mcfg = self.model_config[mod_name]
                if 'model_use_vars' in mcfg and obs_var in mcfg['model_use_vars']:
                    if mcfg['model_use_vars'][obs_var] != mod_var:
                        const.print_log.warning('Ignoring map file {}'
                                                .format(f))
                        continue
                tab.append([obs_var, obs_name, vert_code, mod_name, mod_var])
        return DataFrame(tab, columns=['Var' , 'Obs', 'vert',
                                       'Model', 'Model var'])
    
    def get_obsvar_name_and_type(self, obs_var):
        """Get menu name and type of observation variable
        
        Parameters
        ----------
        obs_var : str
            Name of observation variable
        
        Returns
        -------
        str
            menu name of this variable
        str
            menu category of this variable
        """
        
        try:
            name, tp, cat = self.var_mapping[obs_var]
        except:
            name, tp, cat = obs_var, 'UNDEFINED', 'UNDEFINED'
            self._log.warning('Missing menu name definition for var {}. '
                              'Using variable name'.format(obs_var))
        return (name, tp, cat)
    
    def update_interface(self, **opts):
        """Update web interface
        
        Things done here:
            
            - Update menu file
            - Make web info table json (tab informations in interface)
            - update and order heatmap file
        """
        self.update_menu(**opts)
        self.make_info_table_web()
        self.update_heatmap_json()
        
    def update_menu(self, **opts):
        """Updates menu.json based on existing map json files"""
        update_menu_evaluation_iface(self, **opts)
        
        
    def make_info_table_web(self):
        """Make and safe table with detailed infos about processed data files
        
        The table is stored in as file minfo.json in directory :attr:`exp_dir`.
        """
        table = make_info_table_evaluation_iface(self)
        outname = os.path.join(self.exp_dir, 'minfo.json')       
        with open(outname, 'w+') as f:
            f.write(simplejson.dumps(table, indent=2))
        return table
    
    def _obs_config_asdict(self):
        d = {}
        for k, cfg in self.obs_config.items():
            as_dict = {}
            as_dict.update(**cfg)
            d[k] = as_dict
        return d
    
    def _model_config_asdict(self):
        d = {}
        for k, cfg in self.model_config.items():
            as_dict = {}
            as_dict.update(**cfg)
            d[k] = as_dict
        return d        
    
    def delete_experiment_data(self, base_dir=None, proj_id=None, exp_id=None):
        """Delete all data associated with a certain experiment
        
        Parameters
        ----------
        base_dir : str, optional
            basic output direcory (containg subdirs of all projects)
        proj_name : str, optional
            name of project, if None, then this project is used
        exp_name : str, optional
            name experiment, if None, then this project is used
        """
        from pyaerocom.web.helpers import delete_experiment_data_evaluation_iface
        if proj_id is None:
            proj_id = self.proj_id
        if exp_id is None:
            exp_id = self.exp_id
        if base_dir is None:
            base_dir = self.out_basedir
        delete_experiment_data_evaluation_iface(base_dir, proj_id, exp_id)
        self.update_interface()
        
    def clean_json_files(self, update_interface=False):
        """Checks all existing json files and removes outdated data
        
        This may be relevant when updating a model name or similar.
        """
        
        for file in self.all_map_files:
            (obs_name, obs_var, 
             vert_code, 
             mod_name, mod_var) = self._info_from_map_file(file)
            remove=False
            obs_vars = self._get_valid_obs_vars(obs_name)
        
            if not (obs_name in self.obs_config and
                    mod_name in self.model_config and
                    obs_var in obs_vars):
                remove = True
            
            mcfg = self.model_config[mod_name]
            if 'model_use_vars' in mcfg and obs_var in mcfg['model_use_vars']:
                if not mod_var == mcfg['model_use_vars'][obs_var]:
                    remove=True
                    
            if remove:
                const.print_log.info('Removing outdated map file: {}'.format(file))
                os.remove(os.path.join(self.out_dirs['map'], file))
        for fp in glob.glob('{}/*.json'.format(self.out_dirs['ts'])):
            self._check_clean_ts_file(fp)
        if update_interface:
            self.update_interface()
    
    def _get_valid_obs_vars(self, obs_name):
        if obs_name in self._valid_obs_vars:
            return self._valid_obs_vars[obs_name]
    
        obs_vars = self.obs_config[obs_name]['obs_vars']
        add = []
        for mname, mcfg in self.model_config.items():
            if 'model_add_vars' in mcfg:
                for ovar, mvar in mcfg['model_add_vars'].items():
                    if ovar in obs_vars and not mvar in add:
                        add.append(mvar)
        obs_vars.extend(add)
        self._valid_obs_vars[obs_name]  = obs_vars
        return obs_vars
    
    def _check_clean_ts_file(self, fp):
        spl = os.path.basename(fp).split('OBS-')[-1].split(':')
        obs_name = spl[0]
        obs_var = spl[1].split('_')[0]
        if not (obs_name in self.obs_config and 
                obs_var in self._get_valid_obs_vars(obs_name)):
            const.print_log.info('Removing outdated ts file: {}'.format(fp))
            os.remove(fp)
            return
        try:
            data = read_json(fp)
        except:
            const.print_log.exception('FATAL: detected corrupt json file: {}. '
                                      'Removing file...'.format(fp))
            os.remove(fp)
            return
        if all([x in self.model_config for x in list(data.keys())]):
            return
        data_new = {}
        for mod_name in data.keys():
            if not mod_name in self.model_config:
                const.print_log.info('Removing model {} from {}'
                                .format(mod_name, os.path.basename(fp)))
                continue
            
            data_new[mod_name] = data[mod_name]
        
        write_json(data_new, fp)
            
                
                
        
    def to_dict(self):
        """Convert configuration to dictionary"""
        d = {}
        for key, val in self.__dict__.items():
            if key in self.JSON_CFG_IGNORE:
                continue
            elif isinstance(val, dict):
                if key == 'model_config':
                    sub = self._model_config_asdict()
                elif key == 'obs_config':
                    sub = self._obs_config_asdict()
                else:
                    sub = {}
                    for k, v in val.items():
                        if v is not None:
                            sub[k] = v
                d[key] = sub
            else:
                d[key] = val
        return d
    
    def to_json(self, output_dir):
        """Convert configuration to json ini file"""
        d = self.to_dict()
        out_name = 'cfg_{}_{}.json'.format(self.proj_id, self.exp_id)
        
        save_dict_json(d, os.path.join(output_dir, out_name), indent=3)
        return d
        
    def load_config(self, proj_id, exp_id, config_dir=None):
        """Load configuration json file"""
        if config_dir is None:
            if self.config_dir is not None:
                config_dir = self.config_dir
            else:
                config_dir = '.'
        files = glob.glob('{}/cfg_{}_{}.json'.format(config_dir, proj_id, 
                          exp_id))
        if len(files) == 0:
            raise ValueError('No config file could be found in {} for '
                             'project {} and experiment {}'.format(config_dir, 
                              proj_id, exp_id))
        self.from_json(files[0])
        
    def from_json(self, config_file):
        """Load configuration from json config file"""
        with open(config_file, 'r') as f:
            current = simplejson.load(f)
        self.update(**current)
        
    def __str__(self):
        indent = 2
        _indent_str = indent*' '
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}".format(head, len(head)*"-")
        s += ('\nProject ID: {}'
              '\nEperiment ID: {}'
              '\nExperiment name: {}'
              .format(self.proj_id, self.exp_id, self.exp_name))
        s += '\ncolocation_settings: (will be updated for each run from model_config and obs_config entry)'
        for k, v in self.colocation_settings.items():
            s += '\n{}{}: {}'.format(_indent_str, k, v)
        s += '\n\nobs_config:'
        for k, v in self.obs_config.items():
            s += '\n\n{}{}:'.format(_indent_str, k)
            s = dict_to_str(v, s, indent=indent+2)
        s += '\n\nmodel_config:'
        for k, v in self.model_config.items():
            s += '\n\n{}{}:'.format(_indent_str,  k)
            s = dict_to_str(v, s, indent=indent+2)
    
        return s
    
if __name__ == '__main__':    
    cfg_dir = '/home/jonasg/github/aerocom_evaluation/data_new/config_files/'
    stp = AerocomEvaluation('aerocom', 'PIII-optics', config_dir=cfg_dir)
    #stp.make_info_table_web()
    
    stp.make_regions_json()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Caching class for reading and writing of ungridded data Cache objects
"""
from pyaerocom import const, logger, UngriddedData
from pyaerocom.exceptions import (AerocomConnectionError, CacheReadError,
                                  CacheWriteError)

import glob, os, pickle

# TODO: Write data attribute list contains_vars in header of pickled file and
# check if variables match the request
class CacheHandlerUngridded(object):
    """Interface for reading and writing of cache files
    
    Cache filename mask is 
    
    <dataset_to_read>_<var>.pkl
    
    e.g. EBASMC_scatc550aer.pkl
    
    Attributes
    ----------
    reader : ReadUngriddedBase
        reading class for dataset
    loaded_data : dict
        dictionary containing successfully loaded instances of single variable
        :class:`UngriddedData` objects (keys are variable names)
    """
    __version__ = '1.00'
    #: Directory of cache files
    try:
        CACHE_DIR = const.CACHEDIR
    except:
        CACHE_DIR = None
        logger.exception('Pyaerocom cache directory is not defined')
    #: Cache file header keys that are checked (and required unchanged) when
    #: reading a cache file
    CACHE_HEAD_KEYS = ['pyaerocom_version',
                       'newest_file_in_read_dir',
                       'newest_file_date_in_read_dir',
                       'data_revision', 
                       'reader_version', 
                       'ungridded_data_version', 
                       'cacher_version']
    
    def __init__(self, reader=None, cache_dir=None, **kwargs):
        self._reader = None
        self.reader = reader
        
        self.loaded_data = {}
        
        self._cache_dir = cache_dir
        
    @property
    def reader(self):
        """Instance of reader class"""
        if self._reader is None:
            raise AttributeError('No reader class assigned to cache object')
        return self._reader
    
    @reader.setter
    def reader(self, val):
        from pyaerocom.io import ReadUngriddedBase
        if not isinstance(val, ReadUngriddedBase):
            try:
                val = val.get_reader()
                if not isinstance(val, ReadUngriddedBase):
                    raise TypeError('Invalid input for reader')
            except:
                raise TypeError('Invalid input for reader')
        self._reader = val
        self.loaded_data = {}
        
    @property
    def cache_dir(self):
        """Directory where cached files are stored"""
        if self._cache_dir is not None:
            return self._cache_dir
        if self.CACHE_DIR is None or not os.path.exists(self.CACHE_DIR):
            raise FileNotFoundError('Cache directory does not exist: {}'
                                    .format(self.CACHE_DIR))
        return self.CACHE_DIR
    
    @cache_dir.setter
    def cache_dir(self, val):
        if not isinstance(val, str) or not os.path.exists(val):
            raise FileNotFoundError('Input directory does not exist: {}'
                                    .format(val))
        self._cache_dir = val
        
    @property
    def dataset_to_read(self):
        """Data ID of the associated dataset"""
        return self.reader.dataset_to_read
    
    @property
    def data_dir(self):
        """Data directory of the associated dataset"""
        return self.reader.DATASET_PATH        
        
    def file_name(self, var_name):
        """File name of cache file"""
        name = '_'.join([self.dataset_to_read, var_name])
        return name + '.pkl'
    
    def file_path(self, var_name):
        """File path of cache file"""
        return os.path.join(self.cache_dir, self.file_name(var_name))
    
    def _check_pkl_head_vs_database(self, in_handle):
        current = self.cache_meta_info()
        head = pickle.load(in_handle)
        if not isinstance(head, dict):
            raise CacheReadError('Invalid cache file')
        for k, v in head.items():
            if not k in current:
                raise CacheReadError('Invalid cache header key: {}'.format(k))
            elif not v == current[k]:
                const.print_log.info('{} is outdated (value: {}). Current '
                                     'value: {}'.format(k, v, current[k]))
                return False
        return True
    
    def cache_meta_info(self):
        """Dictionary containing relevant caching meta-info"""
        try:
            newest = max(glob.iglob(os.path.join(self.data_dir, '*')), 
                         key=os.path.getctime)
            newest_date = os.path.getctime(newest)
        except Exception as e:
            raise AerocomConnectionError('Failed to establish connection to '
                                         'data server. Reason: {}'.repr(e))
        d = dict.fromkeys(self.CACHE_HEAD_KEYS)
        from pyaerocom import __version__
        
        d['pyaerocom_version'] = __version__
        d['newest_file_in_read_dir'] = newest
        d['newest_file_date_in_read_dir'] = newest_date
        d['data_revision'] = self.reader.data_revision
        d['reader_version'] = self.reader.__version__
        d['ungridded_data_version'] = UngriddedData.__version__ 
        d['cacher_version'] = self.__version__
        return d
    
    def check_and_load(self, var_name):
        """Check if cache file exists and load
        
        Note
        ----
        If a cache file exists for this database, but cannot be loaded or is
        outdated against pyaerocom updates, then it will be removed (the latter
        only if :attr:`pyaerocom.const.RM_CACHE_OUTDATED` is True).
        
        Returns
        -------
        bool
            True, if cache file exists and could be successfully loaded, else
            False. Note: if import is successful, the corresponding data object
            (instance of :class:`pyaerocom.UngriddedData` can be accessed via
            :attr:`loaded_data'
            
        Raises
        ------
        TypeError
            if cached file is not an instance of :class:`pyaerocom.UngriddedData` 
            class (which should not happen)
        """
        try:
            fp = self.file_path(var_name)
        except FileNotFoundError as e:
            logger.warning(repr(e))
            return False
        
        if not os.path.isfile(fp):
            logger.info('No cache file available for {}, {}'
                        .format(self.dataset_to_read, var_name))
            return False
    
        
        delete_existing = const.RM_CACHE_OUTDATED
                
        in_handle = open(fp, 'rb')
        
        
        try:
            ok = self._check_pkl_head_vs_database(in_handle)
        except Exception as e:
            ok = False
            delete_existing = True
            logger.exception('File error in cached data file {}. File will '
                             'be removed and data reloaded'
                             'Error: {}'.format(fp, repr(e)))
        if not ok:
            # TODO: Should we delete the cache file if it is outdated ???
            logger.info('Aborting reading cache file {}. Aerocom database '
                        'or pyaerocom version has changed compared to '
                        'cached version'
                        .format(self.file_name(var_name)))
            in_handle.close()
            if delete_existing: #something was wrong
                const.print_log.info('Deleting outdated cache file: {}'
                                     .format(fp))
                os.remove(self.file_path(var_name))
            return False
        
        # everything is okay
        data = pickle.load(in_handle)
        if not isinstance(data, UngriddedData):
            raise TypeError('Unexpected data type stored in cache file, need '
                            'instance of UngriddedData, got {}'
                            .format(type(data)))
            
        self.loaded_data[var_name] = data
        logger.info('Successfully loaded data for {} from Cache'
                    .format(self.dataset_to_read))
        return True
    
    def write(self, data, var_name=None):
        """Write single-variable instance of UngriddedData to cache
        
        Parameters
        ----------
        data : UngriddedData
            object containing the data (possibly containing multiple variables)
        var_name : str, optional
            name of variable that is supposed to be stored (only required if
            input `data` contains more than one variable)
        """
        meta = self.cache_meta_info()
        
        if not isinstance(data, UngriddedData):
            raise TypeError('Invalid input, need instance of UngriddedData, '
                            'got {}'.format(type(data)))
        if len(data.contains_datasets) > 1:
            raise CacheWriteError('Input UngriddedData object contains '
                                  'datasets: {}. Can only write single '
                                  'dataset objects'
                                  .format(data.contains_datasets))
        if var_name is None:
            if len(data.contains_vars) > 1:
                raise CacheWriteError('Input UngriddedData object for {} contains '
                                      'more than one variable: {}. Please '
                                      'specify which variable should be '
                                      'cached'
                                      .format(self.reader.data_id,
                                              data.contains_vars))
            var_name = data.contains_vars[0]
        
        elif not var_name in data.contains_vars:
            raise CacheWriteError('Cannot write cache file: variable {} does '
                                  'not exist in input UngriddedData object'
                                  .format(var_name))
            
        if len(data.contains_vars) > 1:
            data = data.extract_var(var_name)
            
        fp = self.file_path(var_name)
        logger.info('Writing cache file: {}'.format(fp))
        success = True
        # OutHandle = gzip.open(c__cache_file, 'wb') # takes too much time
        out_handle = open(fp, 'wb')
        
        try:
            # write cache header
            pickle.dump(meta, out_handle, pickle.HIGHEST_PROTOCOL)
            # write data
            pickle.dump(data, out_handle, pickle.HIGHEST_PROTOCOL)
    
        except Exception as e:
            from pyaerocom import print_log
            print_log.exception('Failed to write cache'.format(repr(e)))
            success=False
        finally:    
            out_handle.close()
            if not success:
                os.remove(self.file_path)
        logger.info('Successfully wrote {} data ({}) to disk!'
                    .format(var_name, self.reader.data_id))
        
    def __str__(self):
        return 'Cache handler for {}'.format(self.reader.data_id)
    
if __name__ == "__main__":
    import pyaerocom as pya
    
    import matplotlib.pyplot as plt
    plt.close('all')

    r = pya.io.ReadAeronetSunV3()
    
    data = r.read(vars_to_retrieve=['od550aer', 'ang4487aer'],
                  file_pattern='Bo*')
    
    aod = data.extract_var('od550aer')
    
    cache = CacheHandlerUngridded(r)
    
    cache.write(aod)
    
    r = pya.io.ReadUngridded()
    r.read('AeronetSunV3Lev2.daily', 
           vars_to_retrieve=['od550aer', 'ang4487aer'],
           file_pattern='Bo*')
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  9 14:14:29 2018
"""
import numpy.testing as npt
import numpy as np
import os
from pyaerocom.test.settings import TEST_RTOL, lustre_unavail
from pyaerocom.io.read_aeronet_sdav3 import ReadAeronetSdaV3
    
@lustre_unavail
def test_load_berlin_AeroSdaV2L2D():
    reader = ReadAeronetSdaV3()
    files = reader.find_in_file_list('*Berlin*')
    assert len(files) == 1
    assert os.path.basename(files[0]) == 'Berlin_FUB.lev30', files[0]

    test_vars = ['od500gt1aer',
                 'od500lt1aer',
                 'od500aer',
                 'ang4487aer',
                 'od550aer',
                 'od550gt1aer',
                 'od550lt1aer']
    
    data = reader.read_file(files[0],
                            vars_to_retrieve=test_vars)                
    
    assert all([x in data for x in test_vars])
    
    # more than 100 timestamps
    assert all([len(data[x]) > 100 for x in test_vars])
    
    assert isinstance(data['dtime'][0], np.datetime64)
    assert data['dtime'][0] == np.datetime64('2014-07-06T12:00:00'), data['dtime'][0]
    
    vals = []
    for var in test_vars:
        
        vals.append(np.nanmean(data[var]))
    
    npt.assert_allclose(vals, 
                        [0.032668, 0.143719, 0.176387, 1.356139, 0.154892, 
                         0.028947, 0.125944], 
                        rtol=TEST_RTOL)
    
if __name__=="__main__":
    test_load_berlin_AeroSdaV2L2D()    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pyaerocom import logger, const
from pyaerocom.mathutils import calc_statistics
from pyaerocom.helpers import to_pandas_timestamp
from pyaerocom.exceptions import (CoordinateError, DataDimensionError, 
                                  DataSourceError, 
                                  NetcdfError, VarNotAvailableError)
from pyaerocom.plot.plotscatter import plot_scatter
from pyaerocom.variable import Variable
from pyaerocom.region import valid_region, Region
import numpy as np
import pandas as pd
import os
import xarray

class ColocatedData(object):
    """Class representing colocated and unified data from two sources
    
    Sources may be instances of :class:`UngriddedData` or 
    :class:`GriddedData` that have been compared to each other. 
    
    Note
    ----
    Currently, it is not foreseen, that this object is instantiated from 
    scratch, but it is rather created in and returned by objects / methods
    that perform colocation.
    The purpose of this object is thus, not the creation of colocated objects,
    but solely the analysis of such data as well as I/O features (e.g. save as
    / read from .nc files, convert to pandas.DataFrame, plot station time 
    series overlays, scatter plots, etc.)
    
    In the current design, such an object comprises 3 dimensions, where the 
    first dimension (depth, index 0) is ALWAYS length 2 and specifies the two
    datasets that were compared
    
    Parameters
    ----------
    data : xarray.DataArray or numpy.ndarray or str, optional
        Colocated data. If str, then it is attempted to be loaded from file.
        Else, it is assumed that data is numpy array and that all further 
        supplementary inputs (e.g. coords, dims) for the 
        instantiation of :class:`DataArray` is provided via **kwargs. 
    ref_data_id : str, optional
        ID of reference data 
    **kwargs
        Additional keyword args that are passed to init of :class:`DataArray` 
        in case input :arg:`data` is numpy array.
        
    Raises
    ------
    IOError
        if init fails
    """
    __version__ = '0.10'
    def __init__(self, data=None, **kwargs):
        self._data = None
        
        if data is not None:
            # check if input is DataArray and if not, try to create instance
            # of DataArray. If this fails, raise Exception
            if isinstance(data, xarray.DataArray):
                self.data = data
            elif isinstance(data, np.ndarray):
                try:
                    data = xarray.DataArray(data, **kwargs)
                except Exception as e:
                    raise IOError('Failed to initiate DataArray from input.\n'
                                  'Error: {}'.format(repr(e)))
                self.data = data
            elif isinstance(data, str):
                self.open(data)
            else:
                raise IOError('Failed to interpret input {}'.format(data))
                    
    
    @property
    def data(self):
        """Data object (instance of :class:`xarray.DataArray`)"""
        if self._data is None:
            raise AttributeError('No data available in this object')
        return self._data
    
    @data.setter
    def data(self, val):
        if not isinstance(val, xarray.DataArray):
            raise IOError('Invalid input for data attribute, need instance '
                             'of xarray.DataArray')
        if self._data is not None:
            logger.warning('Overwriting existing data in ColocatedData object')
        self._data = val
    
    @property
    def name(self):
        """Name of data (should be variable name)"""
        return self.data.name
    
    @name.setter
    def name(self, val):
        self.data.name = val
        
    @property
    def ndim(self):
        """Dimension of data array"""
        return self.data.ndim
    
    @property
    def coords(self):
        """Coordinates of data array"""
        return self.data.coords
    
    @property
    def dims(self):
        """Names of dimensions"""
        return self.data.dims
    
    @property
    def shape(self):
        """Shape of data array"""
        return self.data.shape
    
    @property
    def data_source(self):
        """Coordinate array containing data sources (z-axis)"""
        return self.data.data_source
    
    @property
    def var_name(self):
        """Coordinate array containing data sources (z-axis)"""
        return self.data.var_name
    
    @property
    def longitude(self):
        """Array of longitude coordinates"""
        if not 'longitude' in self.data.coords:
            raise AttributeError('ColocatedData does not include longitude '
                                 'coordinate')
        return self.data.longitude
    
    @property
    def latitude(self):
        """Array of latitude coordinates"""
        if not 'latitude' in self.data.coords:
            raise AttributeError('ColocatedData does not include latitude '
                                 'coordinate')
        return self.data.latitude
    
    
    @property
    def time(self):
        """Array containing time stamps"""
        if not 'time' in self.data.dims:
            raise AttributeError('ColocatedData does not include time'
                                 ' coordinate')
        return self.data.time
            
    
    @property
    def start(self):
        """Start datetime of data"""
        return self.time.values[0]
    
    @property
    def stop(self):
        """Stop datetime of data"""
        return self.time.values[-1]
    
    @property
    def ts_type(self):
        """String specifying temporal resolution of data"""
        if not "ts_type" in self.meta:
            raise ValueError('Colocated data object does not contain '
                             'information about temporal resolution')
        return self.meta['ts_type']
    
    @property
    def units(self):
        """Unit of data"""
        try:
            return self.data.attrs['var_units']
        except KeyError:
            logger.warning('Failed to access unit ColocatedData class (may be an '
                        'old version of data)')
    @property
    def unitstr(self):
        unique = []
        u = self.units
        for val in u:
            if val is None:
                val = 'N/D'
            elif not isinstance(val, str):
                val = str(val)
            if not val in unique:
                unique.append(val)
        return ', '.join(unique)
        
    @property
    def meta(self):
        """Meta data"""
        return self.data.attrs
    
    @property
    def num_grid_points(self):
        """Number of lon / lat grid points that contain data"""
        const.print_log.warning(DeprecationWarning('OLD NAME: please use num_coords'))
        return self.num_coords
    
    @property
    def num_coords(self):
        """Total number of lat/lon coordinates"""
        if not self.check_dimensions():
            raise DataDimensionError('Invalid dimensionality...')
        if 'station_name' in self.coords:
            return len(self.data.station_name)
        
        elif self.ndim == 4:
            if not all([x in self.data.dims for x in ('longitude', 'latitude')]):
                raise AttributeError('Cannot determine grid points. Either '
                                     'longitude or latitude are not contained '
                                     'in 4D data object, which contains the '
                                     'following dimensions: {}'.self.data.dims)
            return len(self.data.longitude) * len(self.data.latitude)
        raise DataDimensionError('Could not infer number of coordinates')
    
    @property
    def num_coords_with_data(self):
        """Number of lat/lon coordinates that contain at least one datapoint
        
        Todo
        ----
        check 4D data 
        """
        if not self.check_dimensions():
            raise DataDimensionError('Invalid dimensionality...')
        
        return (self.data[0].count(dim='time') > 0).data.sum()
        
    def min(self):
        return self.data.min()
    
    def max(self):
        return self.data.max()
    
    def check_dimensions(self):
        """Checks if data source and time dimension are at the right index"""
        dims = self.data.dims
        if not 2 < len(dims) < 5:
            logger.info('Invalid number of dimensions. Must be 3 or 4')
            return False
        try:
            if dims.index('data_source') == 0 and dims.index('time') == 1:
                return True
            raise Exception
        except:
            return False
    
    def resample_time(self, to_ts_type, how='mean',
                      apply_constraints=None, min_num_obs=None, 
                      colocate_time=True, inplace=True, **kwargs):
        """Resample time dimension
        
        Parameters
        ----------
        to_ts_type : str
            new temporal resolution (must be lower than current resolution)
        
        Returns
        -------
        ColocatedData
            new data object containing resampled data
            
        Raises
        ------
        TemporalResolutionError
            if input resolution is higher than current resolution
        """
        if inplace:
            col = self
        else:
            col = self.copy()
            
# =============================================================================
#         to, current = TsType(to_ts_type), TsType(col.ts_type)
#         if to == current:
#             const.print_log.info('Skipping resampling of ColocatedData. Data is '
#                                  'already in {} resolution'.format(to_ts_type))
#             return col
#         elif to > current:
#             from pyaerocom.exceptions import TemporalResolutionError
#             raise TemporalResolutionError('Cannot resample time: input resolution '
#                                           '{} is higher than current {}'
#                                           .format(to_ts_type, current.val))
#         
# =============================================================================
        # if colocate time is activated, remove datapoints from model, where
        # there is no observation
        if colocate_time:
            mask = np.isnan(col.data[0]).data
            col.data.data[1][mask] = np.nan
        # First, compute a mask (in new resolution) that fulfills the 
        # minimum number of measurements requirement
# =============================================================================
#         invalid = None
#         if min_num_obs is not None:
#             pd_freq=to.to_pandas()
#             if not pd_freq in XARR_TIME_GROUPERS:
#                 raise ValueError('Cannot infer xarray grouper for ts_type {}'
#                                  .format(to.val))
#             gr = XARR_TIME_GROUPERS[pd_freq]
#             # 2D mask with shape of resampled data array
#             invalid = col.data.groupby('time.{}'.format(gr)).count(dim='time') < min_num_obs
#             
#         
#             
#         freq = TS_TYPE_TO_PANDAS_FREQ[to.val]
#         
#         data_arr = col.data.resample({'time' : freq}).mean(dim='time')
#         if invalid is not None:
#             data_arr.data[invalid.data] = np.nan
# =============================================================================
        from pyaerocom.time_resampler import TimeResampler
        
        res = TimeResampler(col.data)
        data_arr = res.resample(to_ts_type=to_ts_type,
                                from_ts_type=col.ts_type,
                                how=how,
                                apply_constraints=apply_constraints,
                                min_num_obs=min_num_obs, **kwargs)
        
        data_arr.attrs.update(col.meta)
        data_arr.attrs['ts_type'] = to_ts_type
        
        col.data = data_arr
        col.data.attrs['colocate_time'] = colocate_time
        col.data.attrs.update(res.last_setup)
        
        return col
    
    def get_coords_valid_obs(self):
        
        obs = self.data[0]
        if self.ndim == 4:
            stacked = obs.stack(x=['latitude', 'longitude'])
            invalid = stacked.isnull().all(dim='time')
            coords = stacked.x[~invalid].values
            return list(zip(*list(coords)))
        
        invalid = obs.isnull().all(dim='time')
        return (list(obs.latitude[~invalid].values), 
                list(obs.longitude[~invalid].values))
        
    def copy(self):
        """Copy this object"""
        return ColocatedData(self.data.copy())
    
    def calc_statistics(self, constrain_val_range=False, **kwargs):
        """Calculate statistics from data ensemble
        
        Wrapper for function :func:`pyaerocom.mathutils.calc_statistics` 
        
        Returns
        -------
        dict
            dictionary containing statistical parameters
        """
        if constrain_val_range:
            var = Variable(self.meta['var_name'][1])
            kwargs['lowlim'] = var.lower_limit
            kwargs['highlim'] = var.upper_limit
            
        stats = calc_statistics(self.data.values[1].flatten(),
                                self.data.values[0].flatten(),
                                **kwargs)
        stats['num_coords_with_data'] = self.num_coords_with_data
        stats['num_coords_tot'] = self.num_coords
        return stats
        
    def plot_scatter(self, constrain_val_range=False,  **kwargs):
        """Create scatter plot of data
        
        Parameters
        ----------
        **kwargs
            keyword args passed to :func:`pyaerocom.plot.plotscatter_new.plot_scatter`
            
        Returns
        -------
        ax 
            matplotlib axes instance
        """
        meta = self.meta
        num_points = self.num_coords_with_data
        vars_ = meta['var_name']
        
        if constrain_val_range:
            var = Variable(self.meta['var_name'][1])
            kwargs['lowlim_stats'] = var.lower_limit
            kwargs['highlim_stats'] = var.upper_limit
            
        if vars_[0] != vars_[1]:
            var_ref = vars_[0]
        else:
            var_ref = None
        return plot_scatter(x_vals=self.data.values[0].flatten(), 
                            y_vals=self.data.values[1].flatten(),
                            var_name=vars_[1],
                            var_name_ref = var_ref,
                            x_name=meta['data_source'][0],
                            y_name=meta['data_source'][1], 
                            start=self.start, 
                            stop=self.stop, 
                            unit=self.unitstr,
                            ts_type=meta['ts_type'], 
                            stations_ok=num_points,
                            filter_name=meta['filter_name'], 
                            **kwargs)
    
    def _load_fake_data(self, num_stations=1000, num_tstamps=365):
        raise NotImplementedError('Currently not working')
        data = np.empty((2, num_tstamps, num_stations))
        
        # supposed to be like station coordinate, i.e. the final data type
        lons = np.linspace(10, 40, num_stations)
        lats = np.linspace(30, 50, num_stations)
        
        times = np.linspace(1,30, num_tstamps).astype('datetime64[D]')
        
        times = times.astype('datetime64[ns]')
        
        model = np.ones(num_tstamps).reshape((num_tstamps,1)) * np.linspace(1, 2, num_stations)
        
        obs = np.ones(num_tstamps).reshape((num_tstamps,1))*2 * np.linspace(1, 1.5, num_stations)
        
        data[0] = model
        data[1] = obs
        
        meta = {'ts_type'         : 'daily',
                'ts_type_src'     : '3hourly',
                'year'            : 1970, 
                'data_source'     : ['AeronetSunV2L2.daily', 'ECMWF_OSUITE'],
                'var_name'        : 'od550aer',
                'filter_name'     :  'WORLD-noMOUNTAINS',
                'data_level'      : 'colocated'}
        
        arr = xarray.DataArray(data, coords={'data_source'   : meta['data_source'], 
                                             'time'     : times, 
                                             'longitude': lons,
                                             'latitude' : ('longitude', lats)},
                                dims=['data_source', 'time', 'longitude'],
                                name=meta['var_name'],
                                attrs=meta)
        self._data = arr
    
    def rename_variable(self, var_name, new_var_name, data_source, 
                        inplace=True):
        """Rename a variable in this object
        
        Parameters
        ----------
        var_name : str
            current variable name
        new_var_name : str
            new variable name
        data_source : str
            name of data source (along data_source dimension)
        inplace : bool
            replace here or create new instance
        
        Returns
        -------
        ColocatedData
            instance with renamed variable
            
        Raises
        ------
        VarNotAvailableError
            if input variable is not available in this object
        DataSourceError
            if input data_source is not available in this object
        """
        if not data_source in self.meta['data_source']:
            raise DataSourceError('No such data source {} in ColocatedData'
                                  .format(data_source))
        if not var_name in self.meta['var_name']:
            raise VarNotAvailableError('No such variable {} in ColocatedData'
                                       .format(var_name))
    
        if inplace:
            obj = self
        else:
            obj = self.copy()
        arr = obj.data
        idx = arr.attrs['data_source'].index(data_source)
        arr.attrs['var_name'][idx] = new_var_name
        if var_name == arr.name:
            arr.name = new_var_name
        obj.data = arr
        return obj
        
    @staticmethod
    def _aerocom_savename(var_name, obs_id, model_id, start_str, 
                          stop_str, ts_type, filter_name):
        return ('{}_REF-{}_MOD-{}_{}_{}_{}_{}'.format(var_name,
                                                      obs_id, 
                                                      model_id, 
                                                      start_str, 
                                                      stop_str,
                                                      ts_type,
                                                      filter_name))
    @property
    def savename_aerocom(self):
        """Default save name for data object following AeroCom convention"""
        start_str = self.meta['start_str']
        stop_str = self.meta['stop_str']
        
        source_info = self.meta['data_source']
        data_ref_id = source_info[0]
        if len(source_info) > 2:
            model_id = 'MultiModels'
        else:
            model_id = source_info[1]
        return self._aerocom_savename(self.name,
                                      data_ref_id,
                                      model_id,
                                      start_str,
                                      stop_str,
                                      self.ts_type,
                                      self.meta['filter_name'])
        
    @staticmethod
    def get_meta_from_filename(file_path):
        """Get meta information from file name
        
        Note
        ----
        This does not yet include IDs of model and obs data as these should
        be included in the data anyways (e.g. column names in CSV file) and
        may include the delimiter _ in their name.
        
        Returns 
        -------
        dict
            dicitonary with meta information
        """
        spl = os.path.basename(file_path).split('.nc')[0].split('_')
        
        start = to_pandas_timestamp(spl[-4])
        stop = to_pandas_timestamp(spl[-3])
        
        meta = dict(var_name      = spl[0],
                    ts_type       = spl[-2],  
                    filter_name   = spl[-1],
                    start         = start,
                    stop          = stop)
        
        if not 'REF' in spl[1]:
            raise ValueError('File name does not follow convention')
        ref_base = spl[1].split('REF-')[1]
        mod_base = ''
        in_mod = False
        for item in spl[2:-4]:
            if in_mod:
                mod_base += '_{}'.format(item)
            if item.startswith('MOD-'):
                in_mod=True
                mod_base = item.split('MOD-')[1]
            if not in_mod:
                ref_base += item
        #model, ts_type_src = mod_base.rsplit('-',1)
        meta['data_source'] = [ref_base, mod_base]
        #meta['ts_type_src'] = ts_type_src
        return meta
      
    def to_netcdf(self, out_dir, savename=None, **kwargs):
        """Save data object as .nc file
        
        Wrapper for method :func:`xarray.DataArray.to_netdcf` 
        Parameters
        ----------
        out_dir : str
            output directory
        savename : :obj:`str`, optional
            name of file, if None, the default save name is used (cf. 
            :attr:`savename_aerocom`)
        **kwargs
            additional, optional keyword arguments passed to 
            :func:`xarray.DataArray.to_netdcf` 
            
        """
        if 'path' in kwargs:
            raise IOError('Path needs to be specified using input parameters '
                          'out_dir and savename')
        if savename is None:
            savename = self.savename_aerocom
        if not savename.endswith('.nc'):
            savename = '{}.nc'.format(savename)
        out = None
        for k, v in self.data.attrs.items():
            if v is None:
                self.data.attrs[k] = 'None'
            elif isinstance(v, bool):
                self.data.attrs[k] = int(v)
            if k == 'min_num_obs' and isinstance(v, dict):
                out = ''
                for to, how in v.items():
                    for fr, num in how.items():
                        out += '{},{},{};'.format(to, fr, num)
                
        if out is not None:
            self.data.attrs['_min_num_obs'] = out
            self.data.attrs.pop('min_num_obs')
        self.data.to_netcdf(path=os.path.join(out_dir, savename), **kwargs)
      
    def read_netcdf(self, file_path):
        """Read data from NetCDF file
        
        Parameters
        ----------
        file_path : str
            file path
            
        """
        try:
            self.get_meta_from_filename(file_path)
        except Exception as e:
            raise NetcdfError('Invlid file name for ColocatedData: {}.Error: {}'
                              .format(os.path.basename(file_path, repr(e))))
        arr = xarray.open_dataarray(file_path)
        if '_min_num_obs' in arr.attrs:
            info = {}
            for val in arr.attrs['_min_num_obs'].split(';')[:-1]:
                to, fr, num = val.split(',')
                if not to in info:
                    info[to] = {}
                if not fr in info[to]:
                    info[to][fr] = {}
                info[to][fr] = int(num)
            arr.attrs['min_num_obs'] = info
        self.data = arr
        return self
    
    def to_dataframe(self):
        """Convert this object into pandas.DataFrame
        
        Note
        ----
        This does not include meta information
        """
        logger.warning('This method is currently not completely finished')
        model_vals = self.data.values[1].flatten()
        obs_vals = self.data.values[0].flatten()
        mask = ~np.isnan(obs_vals)
        return pd.DataFrame({'ref'  : obs_vals[mask],
                             'data' : model_vals[mask]})
    
    def from_dataframe(self, df):
        """Create colocated Data object from dataframe
        
        Note
        ----
        This is intended to be used as back-conversion from :func:`to_dataframe`
        and methods that use the latter (e.g. :func:`to_csv`).
        """
        raise NotImplementedError('Coming soon...')
        data = df.to_xarray()
        self.data = data
        self.check_dimensions()
        
        
    def to_csv(self, out_dir, savename=None):
        """Save data object as .csv file
        
        Converts data to pandas.DataFrame and then saves as csv
        
        Parameters
        ----------
        out_dir : str
            output directory
        savename : :obj:`str`, optional
            name of file, if None, the default save name is used (cf. 
            :attr:`savename_aerocom`)
        """
        if savename is None:
            savename = self.savename_aerocom
        if not savename.endswith('.csv'):
            savename = '{}.csv'.format(savename)
        if not self.check_dimensions():
            raise IOError('Invalid dimensionality, please check...')
        df = self.to_dataframe()
        file_path = os.path.join(out_dir, savename)
        df.to_csv(file_path)
        return file_path
        
    def from_csv(self, file_path):
        """Read data from CSV file
        
        Todo
        -----
        Complete docstring
        """
        meta = self.get_meta_from_filename(file_path)
        df = pd.read_csv(file_path)
        self.from_dataframe(df)
        self.data.attrs.update(**meta)
        
    def open(self, file_path):
        """High level helper for reading from supported file sources
        
        Parameters
        ----------
        file_path : str
            file path
        """
        if file_path.endswith('nc'):
            self.read_netcdf(file_path)
            return
        
        raise IOError('Failed to import file {}. File type is not supported '
                      .format(os.path.basename(file_path)))
    
    
        
    def _check_latlon_coords(self):
        _check = ('latitude', 'longitude')
        if not all([x in self.coords for x in _check]):
            raise CoordinateError('Missing latitude or longitude coordinate '
                                  '(or both)')
        elif any([x in self.dims for x in _check]):
            if not all([x in self.dims for x in _check]):
                raise CoordinateError('Only one of latitude / longitude is '
                                      'dimension (require None or both)')
            return False
        return True
        
        
    def _filter_latlon_2d(self, lat_range, lon_range):
        
        if not 'station_name' in self.dims:
            raise DataDimensionError('Cannot filter region, require dimension '
                                      'station_name')
        arr = self.data
        if not list(self.dims).index('station_name') == 2:
            raise DataDimensionError('station_name must be 3. dimensional index')
            
        mask = (np.logical_and(arr.longitude > lon_range[0], 
                               arr.longitude < lon_range[1]) & 
                np.logical_and(arr.latitude > lat_range[0], 
                               arr.latitude < lat_range[1]))
        
        return arr[:,:,mask]
    
    def _filter_latlon_3d(self, lat_range, lon_range):
        if not isinstance(lat_range, slice):
            lat_range = slice(lat_range[0], lat_range[1])
        if not isinstance(lon_range, slice):
            lon_range = slice(lon_range[0], lon_range[1])
            
        return self.data.sel(dict(latitude=lat_range, longitude=lon_range))
    
    def apply_latlon_filter(self, lat_range=None, lon_range=None, 
                            region_id=None):
        """Apply regional filter 
        
        Returns new object filtered for input coordinate range
        
        Parameters
        ----------
        lat_range : list, optional
            latitude range that is supposed to be applied. If specified, then
            also lon_range need to be specified, else, region_id is checked 
            against AeroCom default regions (and used if applicable)
        lon_range : list, optional
            longitude range that is supposed to be applied. If specified, then
            also lat_range need to be specified, else, region_id is checked 
            against AeroCom default regions (and used if applicable)
        region_id : str
            name of region to be applied 
        
        Returns
        -------
        ColocatedData
            filtered data object
        """
        is_2d = self._check_latlon_coords()
        if valid_region(region_id):
            reg = Region(region_id)
            if lon_range is not None and lon_range != reg.lon_range:
                raise ValueError('Conflict: receieved region ID {}, which has '
                                 'a different longitude range ({}) than input '
                                 'lon_range {}'
                                 .format(region_id, reg.lon_range, lon_range))
            if lat_range is not None and lat_range != reg.lat_range:
                raise ValueError('Conflict: receieved region ID {}, which has '
                                 'a different longitude range ({}) than input '
                                 'lat_range {}'
                                 .format(region_id, reg.lat_range, lat_range))
            lon_range = reg.lon_range
            lat_range = reg.lat_range
            region_id = reg.name
            
        if lon_range is None and lat_range is None:
            raise ValueError('Need either lon_range or lat_range or valid '
                             'region_id')
        if lon_range is None:
            lon_range = [-180, 180]
        if lat_range is None:
            lat_range = [-90, 90]
        
        latr, lonr = self.data.attrs['lat_range'], self.data.attrs['lon_range']
        if np.equal(latr, lat_range).all() and np.equal(lonr, lon_range).all():
            const.print_log.info('Filtering of lat_range={} and lon_range={} '
                                 'results in unchanged object, returning self'
                                 .format(lon_range, lat_range))
            return self
        
        if lat_range[0] < latr[0]:
            lat_range[0] = latr[0]
        if lat_range[1] > latr[1]:
            lat_range[1] = latr[1]
        if lon_range[0] < lonr[0]:
            lon_range[0] = lonr[0]
        if lon_range[1] > lonr[1]:
            lon_range[1] = lonr[1]
            
        if is_2d:
            filtered = self._filter_latlon_2d(lat_range, lon_range)
        else:
            filtered = self._filter_latlon_3d(lat_range, lon_range)
            
        if not isinstance(region_id, str):
            region_id = 'CUSTOM'
        try:
            alt_info = filtered.attrs['filter_name'].split('-', 1)[-1]
        except:
            alt_info = 'CUSTOM'
        
        filtered.attrs['filter_name'] = '{}-{}'.format(region_id, alt_info)
        filtered.attrs['region'] = region_id
        filtered.attrs['lon_range'] = lon_range
        filtered.attrs['lat_range'] = lat_range
    
        return ColocatedData(filtered)
        
    def plot_coordinates(self, marker='x', markersize=12, fontsize_base=10, 
                         **kwargs):
        
        from pyaerocom.plot.plotcoordinates import plot_coordinates
        
        lats, lons = self.get_coords_valid_obs()
        return plot_coordinates(lons=lons, 
                                lats=lats, 
                                marker=marker, markersize=markersize,
                                fontsize_base=fontsize_base, **kwargs)
        
    def __contains__(self, val):
        return self.data.__contains__(val)
    
    def __repr__(self):
        return repr(self.data)
    
    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}".format(head, len(head)*"-")
        try:
            data_str = str(self.data)
            s += '\nData: {}'.format(data_str)
        except:
            pass
        return s
    
    ### Deprecated (but still supported) stuff
    @property
    def unit(self):
        """Unit of data"""
        const.print_log.warn(DeprecationWarning('Attr. unit is deprecated, '
                                                'please use units instead'))
        return self.units

if __name__=="__main__":
    
    import matplotlib.pyplot as plt
    import pyaerocom as pya
    plt.close('all')
    
    col_filename = 'absc550aer_REF-EBAS-Lev3_MOD-CAM5-ATRAS_20100101_20101231_daily_WORLD-noMOUNTAINS.nc'
    
    meta = ColocatedData.get_meta_from_filename(col_filename)
    
    obsdata = pya.io.ReadUngridded().read('AeronetSunV3Lev2.daily', 'od550aer')
    modeldata = pya.io.ReadGridded('ECMWF_CAMS_REAN').read_var('od550aer', start=2010)

    coldata1 = pya.colocation.colocate_gridded_ungridded(modeldata, obsdata, 
                                                         ts_type='daily',
                                                         start=2010,
                                                         var_outlier_ranges={'od550aer':[0,10]},
                                                         filter_name='WORLD-noMOUNTAINS',
                                                         remove_outliers=True, 
                                                         colocate_time=False)
    coldata1 = coldata1.resample_time('yearly', colocate_time=True)
    coldata1.plot_coordinates()
    
    sat = pya.io.ReadGridded('MODIS6.aqua').read_var('od550aer', start=2010)
    
    coldata2 = pya.colocation.colocate_gridded_gridded(modeldata, sat, 
                                                       ts_type='monthly',
                                                       regrid_res_deg=10)
    
   
    coldata2.plot_coordinates()
    
    sat_namerica = coldata2.apply_latlon_filter(region_id='NAMERICA')
    sat_namerica.plot_scatter()
    sat_namerica.plot_coordinates()
    
    coldata1.plot_scatter()
    
    stats = {}
    for region_id in pya.region.get_all_default_region_ids():
        filtered = coldata1.apply_latlon_filter(region_id=region_id)
        stats[region_id] = filtered.calc_statistics()
    print('AERONET')    
    for r, s in stats.items():
        if s['num_valid'] == 0:
            print('No data in region {}'.format(r))
        else:
            print('{}: NMB={:.3f} (R={:.2f})'.format(r, s['nmb']*100, s['R']))

    stats = {}
    for region_id in pya.region.get_all_default_region_ids():
        filtered = coldata2.apply_latlon_filter(region_id=region_id)
        stats[region_id] = filtered.calc_statistics()
    print('MODIS')    
    for r, s in stats.items():
        if s['num_valid'] == 0:
            print('No data in region {}'.format(r))
        else:
            print('{}: NMB={:.3f} (R={:.2f})'.format(r, s['nmb']*100, s['R']))
    
    
    
    
        
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  1 09:06:06 2018

@author: jonasg
"""
import re
from os import listdir
from os.path import isdir,join
import fnmatch
from pyaerocom import const, logger
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.exceptions import DataSearchError

class AerocomBrowser(BrowseDict):
    """Interface for browsing all Aerocom data direcories
    
    Note
    ----
    Use :func:`browse` to find directories matching a 
    certain search pattern.
    The class methods :func:`find_matches` and :func:`find_dir` both use 
    :func:`browse`, the only difference is, that the :func:`find_matches` adds
    the search result (a list with strings) to 
    
    """
# =============================================================================
#     def __getitem__(self, name_or_pattern):
#         try:
#             return super(AerocomBrowser, self).__getitem__(name_or_pattern)
#         except KeyError:
#             raise Exception
# =============================================================================
            
    def _browse(self, name_or_pattern, ignorecase=True, return_if_match=True):
        """Search all Aerocom data directories that match input name or pattern
        
        Note
        ----
        Please do not use this function but either 
        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored
        return_if_match : bool
            if True, then the data directory is returned as string, if it can
            be found, else, only a list is returned that contains all 
            matches. The latter takes longer since the whole database is 
            searched.
            
        Returns
        -------
        :obj:`str` or :obj:`list`
            Data directory (str, if ``return_if_match`` is True) or list 
            containing valid Aerocom names (which can then be used to 
            retrieve the paths)
            
        Raises
        ------
        DataSearchError
            if no match or no unique match can be found
        """
        pattern = fnmatch.translate(name_or_pattern)
        _candidates = []
        _msgs = []
        _warnings = []
        
        for obs_id in const.OBS_IDS:
            if ignorecase:
                match = name_or_pattern.lower() == obs_id.lower()
            else:
                match = name_or_pattern == obs_id
            if match:
                logger.info("Found match for search pattern in obs network "
                            "directories {}".format(obs_id))
                self[obs_id] = const.OBSCONFIG[obs_id]["PATH"]
                if return_if_match:
                    return self[obs_id]
            else:
                if ignorecase:
                    match = bool(re.search(pattern, obs_id, re.IGNORECASE))
                else:
                    match = bool(re.search(pattern, obs_id))
            if match:
                self[obs_id] = const.OBSCONFIG[obs_id]["PATH"]
                _candidates.append(obs_id)
                
        for search_dir in const.MODELDIRS:
            # get the directories
            if isdir(search_dir):
                #subdirs = listdir(search_dir)
                subdirs = [x for x in listdir(search_dir) if 
                           isdir(join(search_dir, x))]
                for subdir in subdirs:
                    if ignorecase:
                        match = bool(re.search(pattern, subdir,re.IGNORECASE))
                    else:
                        match = bool(re.search(pattern, subdir))
                    if match:
                        _dir = join(search_dir, subdir)
                        _rnsubdir = join(_dir, "renamed")
                        if isdir(_rnsubdir):
                            logger.info("{} has subdir renamed. Using that one"
                                        .format(_dir))
                            _dir = _rnsubdir
# =============================================================================
#                         ok = True 
#                         if const.GRID_IO.USE_RENAMED_DIR:
#                             logger.info("Checking if renamed directory exists")
#                             _dir = join(_dir, "renamed")
#                             if not isdir(_dir):
#                                 ok = False
#                                 _warnings.append("Renamed folder does not exist "
#                                                  "in {}".format(join(search_dir, 
#                                                      subdir)))
#                         # directory exists and is candidate since it matches 
#                         # the pattern
#                         if ok:
# =============================================================================
                        # append name of candidate ...
                        _candidates.append(subdir)
                        # ... and the corresponding data directory
                        self[subdir] = _dir
                        
                        # now check if it is actually an exact match, if 
                        # applicable
                        if return_if_match:
                        
                            if ignorecase:
                                match = name_or_pattern.lower() == subdir.lower()
                            else:
                                match = name_or_pattern == subdir
                            if match:
                                logger.info("Found match for ID {}"
                                            .format(name_or_pattern))
                                if return_if_match:
                                    return _dir
                                
                                
                        
            else:
                _msgs.append('directory %s does not exist\n'
                                 %search_dir)
        for msg in _msgs:
            logger.info(msg)
            
        for warning in _warnings:
            logger.warning(warning)
        
        if len(_candidates) == 0:
            raise DataSearchError('No matches could be found for search pattern '
                          '{}'.format(name_or_pattern))
        if return_if_match: 
            if len(_candidates) == 1:
                logger.info("Found exactly one match for search pattern "
                            "{}: {}".format(name_or_pattern, _candidates[0]))
                return self[_candidates[0]]
            raise DataSearchError('Found multiple matches for search pattern {}. '
                          'Please choose from {}'.format(name_or_pattern, 
                                              _candidates))
        return _candidates
        
    def find_data_dir(self, name_or_pattern, ignorecase=True):
        """Find match of input name or pattern in Aerocom database
        
        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored
            
        Returns
        -------
        str
            data directory of match
            
        Raises
        ------
        DataSearchError
            if no matches or no unique match can be found
        """
        if name_or_pattern in self:
            logger.info('{} found in instance of AerocomBrowser'.format(name_or_pattern))
            return self[name_or_pattern]
        logger.info('Searching database for {}'.format(name_or_pattern))
        return self._browse(name_or_pattern, ignorecase=ignorecase,
                            return_if_match=True) #returns list
        
    def find_matches(self, name_or_pattern, ignorecase=True):
        """Search all Aerocom data directories that match input name or pattern
        
        Parameters
        ----------
        name_or_pattern : str
            name or pattern of data (can be model or obs data)
        ignorecase : bool
            if True, upper / lower case is ignored
            
        Returns
        -------
        list
            list of names that match the pattern (corresponding paths can be
            accessed from this class instance)
            
        Raises
        ------
        DataSearchError
            if no matches can be found
        """
        return self._browse(name_or_pattern, ignorecase=ignorecase,
                            return_if_match=False) #returns list
        
        
if __name__ == "__main__":
    browser = AerocomBrowser()
    dd = browser.find_data_dir('TM5_AP3-CTRL2016*')
    ea = browser.find_data_dir('Earli*')
    ea1 = browser.find_matches('Earlinet')
    
    print(ea)
    print(ea1)
# =============================================================================
#     try:
#         data_dir = browser.find_data_dir('*Cam5.3-Oslo*')
#     except DataSearchError as e:
#         print(repr(e))
#     
#     
#     
#     matches = browser.find_matches('*Cam5.3-Oslo*')
#     
#     for match in matches:
#         print('{}: {}'.format(match, browser[match]))
#         
#     data_dir_earlinet = browser.find_data_dir('EARLIN*')
#     
#     data_dir_earlinet = browser.find_data_dir('EARLIN*')
#     
#     browser.find_data_dir('Earlinet')
#     
#     
#     
#         
# =============================================================================



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 12 14:45:43 2018

@author: jonasg
"""
import pytest
import numpy as np
import numpy.testing as npt
from pyaerocom.test import settings
import pyaerocom.exceptions as exc
from cf_units import Unit

RTOL = settings.TEST_RTOL

@pytest.fixture
def stat1():
    return settings.DATA_ACCESS['station_data1']

@pytest.fixture
def stat2():
    return settings.DATA_ACCESS['station_data2']

def test_default_vert_grid(stat1):
    grid = stat1.default_vert_grid
    assert grid.mean() == 7375 #m
    step = np.unique(np.diff(grid))
    assert len(step) == 1
    assert step[0] == 250

def test_has_var(stat1, stat2):
    assert stat1.has_var('od550aer')
    assert stat2.has_var('conco3')
    
def test_get_unit(stat1, stat2):
    
    assert stat1.get_unit('ec550aer') == Unit('m-1')
    assert stat2.get_unit('ec550aer') == Unit('1/Mm')
    
def test_check_var_unit_aerocom(stat1):
    assert stat1.get_unit('ec550aer') == Unit('m-1')
    stat1.check_var_unit_aerocom('ec550aer')
    assert stat1.get_unit('ec550aer') == Unit('1/Mm')
    

    
if __name__=="__main__":
    STAT1 = settings.DATA_ACCESS['station_data1']
    STAT2 = settings.DATA_ACCESS['station_data2']
    
    stat1 = STAT1
    stat2 = STAT2
    
    ec_mean =  np.mean(stat1.ec550aer)
    try:
        stat1.check_unit('ec550aer')
    except exc.DataUnitError:
        from pyaerocom import Variable
        stat1.convert_unit('ec550aer', Variable('ec550aer').units)
    
    from pyaerocom.units_helpers import unit_conversion_fac
    
    fac = unit_conversion_fac('m-1', '1/Mm')
    print(stat1)
    
    npt.assert_allclose(ec_mean * fac,
                        stat1.ec550aer.mean(), 
                        rtol=RTOL)
    
    merged = stat1.merge_other(stat2, 'ec550aer')
    print(merged)
    
    
    


def _init_logger():
    import logging
    ### LOGGING
# Note: configuration will be propagated to all child modules of
# pyaerocom, for details see 
# http://eric.themoritzfamily.com/learning-python-logging.html
    logger = logging.getLogger('pyaerocom')
    
    default_formatter = logging.Formatter(\
       "%(asctime)s:%(levelname)s:\n%(message)s")
    
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(default_formatter)
    
    logger.addHandler(console_handler)
    
    logger.setLevel(logging.CRITICAL)
    
    print_log = logging.getLogger('pyaerocom_print')
    
    print_handler = logging.StreamHandler()
    print_handler.setFormatter(logging.Formatter("%(message)s"))
    
    print_log.addHandler(print_handler)
    
    print_log.setLevel(logging.INFO)
    return (logger, print_log)

def change_verbosity(new_level='debug', log=None):
    if log is None:
        log = logger
    if isinstance(new_level, str):
        if not new_level in LOGLEVELS:
            raise ValueError("Invalid input for loglevel, choose "
                             "from {}".format(LOGLEVELS.keys()))
        new_level = LOGLEVELS[new_level]
    log.setLevel(new_level)
    
### Functions for package initialisation
def _init_supplemental():
    from pkg_resources import get_distribution
    from os.path import abspath, dirname
    return (get_distribution('pyaerocom').version, abspath(dirname(__file__)))

def check_requirements(logger):
    GEONUM_AVAILABLE = True
    BASEMAP_AVAILABLE = True
    try:
        import geonum
    except:
        GEONUM_AVAILABLE = False
        logger.warning('geonum library is not installed. Some features will not '
                    'be available (e.g. conversion of pressure to altitude')
    try:
        from mpl_toolkits.basemap import Basemap
    except:
        BASEMAP_AVAILABLE = False
        logger.warning('basemap extension library is not installed (or cannot be '
                    'imported. Some features will not be available')
    return (GEONUM_AVAILABLE, BASEMAP_AVAILABLE)

__version__, __dir__ = _init_supplemental()

logger, print_log = _init_logger()
LOGLEVELS = {'debug': 10,
             'info': 20,
             'warning': 30,
             'error': 40,
             'critical': 50}

(GEONUM_AVAILABLE, 
 BASEMAP_AVAILABLE) = check_requirements(logger)

# Imports
from . import _lowlevel_helpers
from . import obs_io
# custom toplevel classes
from .variable import Variable
from .region import Region
from .config import Config

const = Config()
    
from . import metastandards
from . import mathutils
from . import geodesy
from . import vert_coords

from .vertical_profile import VerticalProfile
from .stationdata import StationData
from .griddeddata import GriddedData
from .ungriddeddata import UngriddedData
from .filter import Filter
from .colocateddata import ColocatedData
from .colocation_auto import ColocationSetup, Colocator
from . import colocation
#from . import analysis

from . import io
from . import plot
from . import interactive
from . import web


#from .ungriddeddata import UngriddedData
from .io.helpers import search_data_dir_aerocom
from .io.utils import browse_database
from .variable import get_variable
from .utils import create_varinfo_table
#from .obsdata import ObsData, ProfileData, StationData
#print('Elapsed time init pyaerocom: {} s'.format(time()-t0))






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
from pyaerocom.plot.mapping import init_map
import cartopy

def plot_coordinates(lons, lats, xlim=(-180, 180), ylim=(-90, 90), 
                     label=None, legend=True, color='r', marker='o', 
                     markersize=8, fontsize_base=10,
                     ax=None, **kwargs):
    """Plot input coordinates on a map
    
    lons : :obj:`ndarray`
        array of longitude coordinates (can also be list or tuple)
    lats : :obj:`ndarray`
        array of latitude coordinates (can also be list or tuple)
    
    
    """
    if not isinstance(ax, cartopy.mpl.geoaxes.GeoAxes):
        ax = init_map(xlim, ylim, ax=ax, **kwargs)
    
    if label is None:
        label = '{} stations'.format(len(lons))
    else:
        label += ' ({} stations)'.format(len(lons))
    ax.scatter(lons, lats, markersize, marker=marker, color=color,
               label=label)
    
    if legend and label:
        ax.legend()
    
    return ax

if __name__ == '__main__':
    import pyaerocom as pya
    import matplotlib.pyplot as plt
    plt.close('all')
    r = pya.io.ReadUngridded()
    
    d = r.read(pya.const.AERONET_SUN_V3L2_AOD_DAILY_NAME, 
               vars_to_retrieve=['od550aer'])
    
    d1 = r.read('EBASMC', vars_to_retrieve=['scatc550aer'])
    
    ax = plot_coordinates(d.longitude, d.latitude, label='AERONET L2V3',
                          legend=False)
    
    
    ax = plot_coordinates(d1.longitude, d1.latitude, label='EBAS',
                          legend=True, ax=ax,
                          color='lime', markersize=22, marker='^')
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  9 14:14:29 2018
"""
import numpy.testing as npt
import numpy as np
import os
from pyaerocom.test.settings import TEST_RTOL, lustre_unavail
from pyaerocom.io.read_earlinet import ReadEarlinet
from pyaerocom import VerticalProfile

TESTDIR = '/lustre/storeA/project/aerocom/aerocom1/AEROCOM_OBSDATA/Export/Earlinet/CAMS/data/ev/'
FILES = ['ev1008192050.e532',
         'ev1009162031.e532',
         'ev1012131839.e532',
         'ev1011221924.e532',
         'ev1105122027.e532']

PATHS = [os.path.join(TESTDIR, f) for f in FILES]

@lustre_unavail
def test_all_files_exist():
    for file in PATHS:
        if not os.path.exists(file):
            raise AssertionError('File {} does not exist'.format(file))

@lustre_unavail
def test_first_file():
    read = ReadEarlinet()
    read.files = PATHS
    
    stat = read.read_file(PATHS[0], 'ec532aer')
    
    assert 'data_level' in stat
    assert 'wavelength_det' in stat
    assert 'has_zdust' in stat
    assert 'eval_method' in stat
    
    assert 'ec532aer' in stat.var_info
    
    i = stat.var_info['ec532aer']
    assert i['unit_ok']
    assert i['err_read']
    assert i['outliers_removed']
    
    assert isinstance(stat.ec532aer, VerticalProfile)
    
    p = stat.ec532aer
    
    vals_data = [np.nanmean(p.data), np.nanstd(p.data), np.sum(np.isnan(p.data)),
                 len(p.data)]
    vals_dataerr = [np.nanmean(p.data_err), np.nanstd(p.data_err)]
    vals_altitude = [np.min(p.altitude), np.max(p.altitude)]
    
    npt.assert_allclose(vals_data,[4.463068618148296, 1.8529271228530515, 216, 253],
                        rtol=TEST_RTOL)
    npt.assert_allclose(vals_dataerr, [4.49097234883772, 0.8332285038985179],
                        rtol=TEST_RTOL)
    npt.assert_allclose(vals_altitude, [331.29290771484375, 7862.52490234375],
                        rtol=TEST_RTOL)


@lustre_unavail
def test_read_ungridded():
    read = ReadEarlinet()
    read.files = PATHS
    
    data = read.read('ec532aer')
    
    npt.assert_equal(len(data.metadata), 5)
    npt.assert_array_equal(data.shape, (786, 12))
    
    npt.assert_allclose([np.nanmin(data._data[:, data._DATAINDEX]),
                         np.nanmean(data._data[:, data._DATAINDEX]),
                         np.nanmax(data._data[:, data._DATAINDEX])],
                        [-0.440742, 24.793547, 167.90787], rtol=TEST_RTOL)
    
    merged = data.to_station_data('Evora', freq='monthly')
    
    npt.assert_allclose([float(np.nanmin(merged.ec532aer)),
                         float(np.nanmean(merged.ec532aer)),
                         float(np.nanmax(merged.ec532aer))],
                        [0.220322,  23.093238, 111.478665],
                        rtol=TEST_RTOL)
    
if __name__=="__main__":

    test_all_files_exist()
    test_first_file()
    test_read_ungridded()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  9 14:14:29 2018
"""
from pyaerocom.test.settings import lustre_unavail
from pyaerocom.io import ReadUngridded

def test_supported():
    supported_datasets =ReadUngridded().supported_datasets
    assert len(supported_datasets) == 17
    assert supported_datasets == ['AeronetInvV3Lev2.daily',
                                  'AeronetInvV3Lev1.5.daily',
                                  'AeronetInvV2Lev2.daily',
                                  'AeronetInvV2Lev1.5.daily',
                                  'AeronetSDAV2Lev2.daily',
                                  'AeronetSDAV3Lev1.5.daily',
                                  'AeronetSDAV3Lev2.daily',
                                  'AeronetSunV2Lev2.daily',
                                  'AeronetSunV2Lev2.AP',
                                  'AeronetSunV3Lev1.5.daily',
                                  'AeronetSunV3Lev1.5.AP',
                                  'AeronetSunV3Lev2.daily',
                                  'AeronetSunV3Lev2.AP',
                                  'EARLINET',
                                  'EBASMC',
                                  'DMS_AMS_CVO',
                                  'GAWTADsubsetAasEtAl']

@lustre_unavail
def test_basic_attributes():
    r = ReadUngridded()
    assert not r.ignore_cache
    assert r.datasets_to_read == ['AeronetSunV3Lev2.daily']
    assert type(r.get_reader()).__name__ == 'ReadAeronetSunV3'
    assert r.dataset_provides_variables() == ['od340aer', 'od440aer',
                                            'od500aer', 'od870aer',
                                            'ang4487aer', 'ang4487aer_calc',
                                            'od550aer']
    
@lustre_unavail
def test_read_aeronet_sunv3():
    r = ReadUngridded()
    data = r.read('AeronetSunV3Lev2.daily',
                   vars_to_retrieve=['od550aer', 'ang4487aer'],
                   file_pattern='Bo*')
    data._check_index()

if __name__=="__main__":

    test_supported()
    test_read_aeronet_sunv3()


    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
I/O helper methods of the pyaerocom package
"""
from datetime import datetime
from pyaerocom import const
from pyaerocom.io import AerocomBrowser
from pyaerocom import __dir__
import os
from pyaerocom.exceptions import (VarNotAvailableError, VariableDefinitionError)


from collections import OrderedDict as od

TSTR_TO_NP_DT = {"hourly"  :  "datetime64[h]",
                 "3hourly" :  "datetime64[3h]",
                 "daily"   :  "datetime64[D]",
                 "monthly" :  "datetime64[M]"}

TSTR_TO_NP_TD = {"hourly"  :  "timedelta64[h]",
                 "3hourly" :  "timedelta64[3h]",
                 "daily"   :  "timedelta64[D]",
                 "monthly" :  "timedelta64[M]"}


TSTR_TO_CF = {"hourly"  :  "hours",
              "3hourly" :  "hours",
              "daily"   :  "days",
              "monthly" :  "days"}

def _print_read_info(i, mod, tot_num, last_t, name, logger):
    """Helper for displaying standardised output in reading classes
    
    Not to be used directly
    """
    t = datetime.now()
    logger.info("Reading files {}-{} of {} ({}) | {} (delta = {} s')"
                .format(i+1,i+1+mod, tot_num, name,
                        t.strftime('%H:%M:%S'),
                        (t-last_t).seconds))
    return t

def get_metadata_from_filename(filename):
    """Try access metadata information from filename"""
    from pyaerocom.io.fileconventions import FileConventionRead
    fc = FileConventionRead().from_file(filename)
    return fc.get_info_from_file(filename)
    
def read_ebas_flags_file(ebas_flags_csv):
    """Reads file ebas_flags.csv
    
    Parameters
    ----------
    ebas_flags_csv : str
        file containing flag info
        
    Returns
    -------
    dict
        dict with loaded flag info
    """
    from pyaerocom._lowlevel_helpers import BrowseDict
    valid = BrowseDict()
    values = BrowseDict()
    info = BrowseDict()
    with open(ebas_flags_csv) as fio:
        for line in fio:
            spl = line.strip().split(',')
            num = int(spl[0].strip())
            try:
                val_str = spl[-1][1:-1]
            except:
                raise IOError('Failed to read flag information in row {} '
                              '(Check if entries in ebas_flags.csv are quoted)'
                              .format(line))
            info_str = ','.join(spl[1:-1])
            try:
                info_str = info_str[1:-1]
            except:
                raise IOError('Failed to read flag information in row {} '
                              '(Check if entries in ebas_flags.csv are quoted)'
                              .format(line))
            isvalid = True if val_str == 'V' else False
            valid[num] = isvalid
            values[num] = val_str
            info[num] = info_str
    result = BrowseDict()
    result.valid = valid
    result.info = info
    result.vals = values
    return result

def add_file_to_log(filepath, err_msg):
    
    try:
        dirname = os.path.dirname(filepath)
        spl = dirname.split(os.sep)
        if spl[-1].lower() == 'renamed':
            model_or_obs_id = spl[-2]
        else:
            model_or_obs_id = spl[-1]
    except:
        model_or_obs_id = 'others'
    try:
        logdir = const.LOGFILESDIR
        found = False
        logfile = os.path.join(logdir, model_or_obs_id + '.log')
        if os.path.exists(logfile):
            with open(logfile, 'r') as f:
                for line in f:
                    if filepath == line.strip():
                        found = True
                        break
            
        if not found:
            with open(logfile, 'a+') as f:
                f.write(filepath + '\n')
            with open(os.path.join(logdir, model_or_obs_id + '_ERR.log'), 'a+') as ferr:
                ferr.write('{}\n{}\n\n'.format(filepath,
                                               err_msg))
    except Exception as e:
        from pyaerocom import print_log
        const.WRITE_FILEIO_ERR_LOG = False
        print_log.info('Failed to write to file-read error logging ({}). '
                       'Deactiving lgging'.format(repr(e)))
        
def get_standard_name(var_name):
    """Get standard name of aerocom variable
    
    Parameters
    ----------
    var_name : str
        HTAP2 variable name
    
    Returns
    --------
    str
        corresponding standard name
        
    Raises
    ------
    VarNotAvailableError
        if input variable is not defined in *variables.ini* file
    VariableDefinitionError
        if standarad name is not set for variable in *variables.ini* file
    """
    if not var_name in const.VARS:
        raise VarNotAvailableError('No such variable {}. Check variables.ini'.format(var_name))
    name = const.VARS[var_name].standard_name
    if name is None:
        raise VariableDefinitionError('standard_name not defined for variable')
    return name

def search_data_dir_aerocom(name_or_pattern, ignorecase=True):
    """Search Aerocom data directory based on model / data ID
    """
    browser = AerocomBrowser()
    return browser.find_data_dir(name_or_pattern, ignorecase)
  
def get_all_supported_ids_ungridded():
    """Get list of datasets that are supported by :class:`ReadUngridded`
    
    Returns
    -------
    list 
        list with supported network names
    """
    from pyaerocom.io import ReadUngridded
    return ReadUngridded().SUPPORTED_DATASETS

def get_obsnetwork_dir(obs_id):
    """Returns data path for obsnetwork ID
    
    Parameters
    ----------
    obs_id : str
        ID  of obsnetwork (e.g. AeronetSunV2Lev2.daily)
        
    Returns
    -------
    str
        corresponding directory from ``pyaerocom.const``
        
    Raises
    ------
    ValueError
        if obs_id is invalid
    IOError
        if directory does not exist
    """
    if not obs_id in const.OBS_IDS:
        raise ValueError("Observation network ID {} does not exist".format(obs_id))
        
    data_dir = const.OBSCONFIG[obs_id]['PATH']
    if not os.path.exists(data_dir):
        raise IOError("Data directory {} of observation network {} does not "
                      "exists".format(data_dir, obs_id))
    return data_dir

def save_dict_json(d, fp, ignore_nan=True, indent=None):
    """Save a dictionary as json file using :func:`simplejson.dump`
    
    Parameters
    ----------
    d : dict
        input dictionary
    fp : str
        filepath of json file
        
    """
    import simplejson
    with open(fp, 'w') as f:
        simplejson.dump(d, f, ignore_nan=ignore_nan, indent=indent)
        
def search_names(update_inifile=True, check_nc_file=True):
    """Search model IDs in database
    
    Parameters
    ----------
    update_inifile : bool
        if True, the file *names.txt* will be updated. The file is located
        in the installation *data* directory.
    check_nc_file : bool
        If True, only model IDs are included, for which at least one nc file
        can be detected in the corresponding renamed sub directory
    """
    names = []
    for mdir in const.MODELDIRS:
        print("\n%s\n" %mdir)
        sub = os.listdir(mdir)
        for item in sub:
            path = os.path.join(mdir, item, "renamed")
            if os.path.isdir(path):
                print("\n%s\n" %path)
                add = True
                if check_nc_file:
                    add = False
                    for name in os.listdir(path):
                        if name.endswith(".nc"):
                            add = True
                            break
                if add:
                    names.append(item)
    names = sorted(od.fromkeys(names))
    if update_inifile:
        fpath = os.path.join(__dir__, "data", "names.txt")
        f = open(fpath, "w") 
        for name in names:
            f.write("%s\n" %name)
        f.close()
    return names

def get_all_names():
    """Try to import all model IDs from file names.txt in data directory"""
    try:
        with open(os.path.join(__dir__, "data", "names.txt")) as f:
            names = f.read().splitlines()
        f.close()
    except:
        try:
            names = search_names()
        except:
            raise Exception("Failed to access model IDs")
    return names
    
if __name__=="__main__":
    #names = search_names()
    names = get_all_names()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
High level test methods that check EBAS time series data for selected stations
Created on Thu Apr 12 14:45:43 2018

@author: jonasg
"""
import pytest
import numpy.testing as npt
import numpy as np
from pyaerocom.test.settings import lustre_unavail
from pyaerocom.io import ReadEbas

def _make_data():
    r = ReadEbas()
    return r.read('scatc550aer', station_names='Jungfrau*')

@pytest.fixture(scope='module')
def data_scat_jungfraujoch():
    return _make_data()

@lustre_unavail
def test_ungriddeddata_jungfraujoch(data_scat_jungfraujoch):
    data = data_scat_jungfraujoch
    
    assert data.shape == (245448, 12)# (227928, 12)
    assert len(data.metadata) == 28
    
    unique_coords = []
    unique_coords.extend(np.unique(data.latitude))
    unique_coords.extend(np.unique(data.longitude))
    unique_coords.extend(np.unique(data.altitude))
    assert len(unique_coords) == 4
    npt.assert_allclose(unique_coords, [46.5475, 7.985, 3578.0, 3580.0],
                        rtol=1e-2)
    
    vals = data._data[:, data.index['data']]
    check = [np.nanmean(vals), 
             np.nanstd(vals),
             np.nanmax(vals),
             np.nanmin(vals)]

    npt.assert_allclose(check, 
                        [4.35, 7.24, 182.7, -5.57], rtol=1e-1)
    
    

@lustre_unavail   
def test_scat_jungfraujoch(data_scat_jungfraujoch):
    stat = data_scat_jungfraujoch.to_station_data('Jung*')
    
    assert 'scatc550aer' in stat
    assert 'scatc550aer' in stat.overlap
    assert len(stat.overlap['scatc550aer']) == 25787 #17466
    assert stat['stat_merge_pref_attr'] == 'revision_date'
    assert int(stat['data_level']) == 2
    
    npt.assert_array_equal([stat.dtime.min(), stat.dtime.max()],
                            [np.datetime64('1995-07-08T23:00:00'), 
                             np.datetime64('2018-12-31T23:00:00')])
    
    vals = [stat['instrument_name'], stat['ts_type'], stat['PI'],
            len(stat.filename.split(';'))]
    
    npt.assert_array_equal(vals,
                           ['Ecotech_Aurora3000_JFJ_dry; TSI_3563_JFJ_dry; IN3563',
                            'hourly',
                            'Bukowiecki, Nicolas; Baltensperger, Urs; Weingartner, Ernest',
                            28])
    
    d = stat.scatc550aer
    vals = [d.mean(), d.std(), d.min(), d.max()]
    npt.assert_allclose(vals, [4.66, 7.60, -5.57, 182.7], rtol=1e-1)
    
    d = stat.overlap['scatc550aer']
    vals = [d.mean(), d.std(), d.min(), d.max()]
    npt.assert_allclose(vals,
                        [2.91, 4.58, -3.47, 97.65], rtol=1e-1)

@lustre_unavail   
def test_scat_jungfraujoch_lev3(data_scat_jungfraujoch):
    data_scat_jungfraujoch = data_scat_jungfraujoch.apply_filters(
            set_flags_nan=True, remove_outliers=True, data_level=2)
    stat = data_scat_jungfraujoch.to_station_data('Jung*')
    
    assert 'scatc550aer' in stat
    assert 'scatc550aer' in stat.overlap
    assert len(stat.overlap['scatc550aer']) == 25787 #17466
    assert stat['stat_merge_pref_attr'] == 'revision_date'
    assert int(stat['data_level']) == 2
    
    npt.assert_array_equal([stat.dtime.min(), stat.dtime.max()],
                            [np.datetime64('1995-07-08T23:00:00'), 
                             np.datetime64('2018-12-31T23:00:00')])
    
    vals = [stat['instrument_name'], stat['ts_type'], stat['PI'],
            len(stat.filename.split(';'))]
    
    npt.assert_array_equal(vals,
                           ['Ecotech_Aurora3000_JFJ_dry; TSI_3563_JFJ_dry; IN3563',
                            'hourly',
                            'Bukowiecki, Nicolas; Baltensperger, Urs; Weingartner, Ernest',
                            28])
    
    d = stat.scatc550aer
    vals = [d.mean(), d.std(), d.min(), d.max()]
    npt.assert_allclose(vals, [4.66, 7.60, -5.57, 182.7], rtol=1e-1)
    
    d = stat.overlap['scatc550aer']
    vals = [d.mean(), d.std(), d.min(), d.max()]
    npt.assert_allclose(vals,
                        [2.91, 4.58, -3.47, 97.65], rtol=1e-1)
    
@lustre_unavail   
def test_scat_jungfraujoch_subset(data_scat_jungfraujoch):
    
    stat = data_scat_jungfraujoch.to_station_data('Jung*', 
                                                  start=2008, stop=2011, 
                                                  freq='monthly')
    
    assert stat.dtime.min() == np.datetime64('2008-01-15')
    assert stat.dtime.max() == np.datetime64('2010-12-15')
    assert stat.ts_type == 'monthly'
    assert stat.ts_type_src == 'hourly'
    
    d = stat['scatc550aer']
    vals = [d.mean(), d.std(), d.min(), d.max()]

    npt.assert_allclose(vals, [4.37, 3.48, 0.46, 11.93], rtol=1e-1)
    
if __name__=="__main__":
   # pya.change_verbosity('info')
    #import sys
    d = _make_data()
    test_ungriddeddata_jungfraujoch(d)
    test_scat_jungfraujoch(d)
    test_scat_jungfraujoch_lev3(d)
    
    test_scat_jungfraujoch_subset(d)
# =============================================================================
#     
#     d = _make_data()
#     stat = d.to_station_data('Jung*', start=2008, stop=2011, freq='monthly')
#     stat.plot_timeseries('scatc550aer')
#     
# =============================================================================
    #pytest.main(sys.argv)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 15 11:05:50 2019

@author: paulinast
"""

import pytest
import numpy.testing as npt
import numpy as np
from pyaerocom.test.settings import lustre_unavail, TEST_RTOL, test_not_working
from pyaerocom.io import ReadGAW

def _make_data():
    r = ReadGAW()
    return r.read('vmrdms')

@pytest.fixture(scope='module')
@lustre_unavail
def data_vmrdms_ams_cvo():
    return _make_data()

@lustre_unavail
@test_not_working
def test_ungriddeddata_ams_cvo(data_vmrdms_ams_cvo):
    data = data_vmrdms_ams_cvo
    # assert data.data_revision['DMS_AMS_CVO'] == 'n/a'
    assert data.shape == (819+7977, 12)
    assert len(data.metadata) == 2
    
    unique_coords = []
    unique_coords.extend(np.unique(data.latitude))
    unique_coords.extend(np.unique(data.longitude))
    unique_coords.extend(np.unique(data.altitude))
    assert len(unique_coords) == 6
    npt.assert_allclose(unique_coords, [-37.8, 16.848, -24.871, 77.53, 10.0, 65.0],
                        rtol=TEST_RTOL)
    
    vals = data._data[:, data.index['data']]
    check = [np.nanmean(vals), 
             np.nanstd(vals),
             np.nanmax(vals),
             np.nanmin(vals)]
    print(check)
    #npt.assert_allclose(vals,
     #                    [174.8499921813917, 233.0328306938496, 2807.6, 0.0],
      #                   rtol=TEST_RTOL)
    
    
@lustre_unavail   
@test_not_working
def test_vmrdms_ams(data_vmrdms_ams_cvo):
    stat = data_vmrdms_ams_cvo.to_station_data(meta_idx= 0)
    
    keys = list(stat.keys())
    assert 'vmrdms' in keys
    assert 'var_info' in keys
    
    npt.assert_array_equal([stat.dtime.min(), stat.dtime.max()],
                            [np.datetime64('1987-03-01T00:00:00.000000000'), 
                             np.datetime64('2008-12-31T00:00:00.000000000')])
    
    vals = [stat['instrument_name'], stat['ts_type'], stat['filename']]
    
    npt.assert_array_equal(vals,
                           ['unknown', 
                            'daily',  
                            'ams137s00.lsce.as.fl.dimethylsulfide.nl.da.dat'])
    
    d = stat['vmrdms']
    vals = [d.mean(), d.std(), d.max(), d.min()]
    npt.assert_allclose(vals,
                         [185.6800736155262, 237.1293922258991, 2807.6, 5.1],
                         rtol=TEST_RTOL)
   
    
@lustre_unavail   
def test_vmrdms_ams_subset(data_vmrdms_ams_cvo):
    
    stat = data_vmrdms_ams_cvo.to_station_data(meta_idx= 0, 
                                               start=2000, stop=2008, 
                                               freq='monthly')
    
    npt.assert_array_equal([stat.dtime.min(), stat.dtime.max()],
                            [np.datetime64('2000-01-15T00:00:00.000000000'),
                             np.datetime64('2007-12-15T00:00:00.000000000') ])
    assert stat.ts_type == 'monthly'
    assert stat.ts_type_src == 'daily'
    
    
if __name__=="__main__":

    d = _make_data()
    test_ungriddeddata_ams_cvo(d)
    test_vmrdms_ams(d)
    test_vmrdms_ams_subset(d)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  4 10:36:15 2019

@author: jonasg
"""
from pyaerocom import geodesy
from pyaerocom.test.settings import geonum_unavail, etopo1_unavail
import numpy.testing as npt


TEST_LAT = 50.8
TEST_LON = 9


def test_haversine():
    npt.assert_allclose(geodesy.haversine(0, 15, 0, 16), 111.2, atol=0.1)
    
def test_is_within_radius_km():
    assert geodesy.is_within_radius_km(0, 15, 0, 16, 1000, 111.2)
    
@geonum_unavail
def test_srtm_altitude():
    npt.assert_almost_equal(geodesy.get_topo_altitude(TEST_LAT, TEST_LON), 207)
    
@geonum_unavail
@etopo1_unavail
def test_etopo_altitude():
    npt.assert_almost_equal(geodesy.get_topo_altitude(TEST_LAT, TEST_LON,
                                                      topo_dataset='etopo1'), 
                            217)
    
if __name__ == '__main__':

    test_haversine()
    test_is_within_radius_km()
    
    test_srtm_altitude()
    test_etopo_altitude()